let isCrawling = false;

chrome.runtime.onMessage.addListener(async (msg, sender) => {
  if (msg.action === "startCrawling") {
    isCrawling = true;
    await startNextTask();
  } else if (msg.action === "stopCrawling") {
    isCrawling = false;
  } else if (msg.action === "taskCompleted") {
    if (!sender.tab?.id) {
      return;
    }
    chrome.tabs.remove(sender.tab.id);
    if (isCrawling) {
      await startNextTask();
    }
  } else if (msg.action === "taskFinished") {
    const values = {
      task: msg.task,
      result: msg.result,
    };
    const resp = await sendToServer("/finish_task", values);
    if (resp === null) {
      isCrawling = false;
      return;
    }
  } else if (msg.action === "expectNavigation") {
    if (!sender.tab?.id) {
      return;
    }
    if (!isCrawling) {
      chrome.tabs.remove(sender.tab.id);
      return;
    }
    const task = msg.task;
    const page = msg.page;

    await new Promise((r) => setTimeout(r, 1000));
    const tabID = sender.tab.id;
    while ((await chrome.tabs.get(tabID)).status !== "complete") {
      await new Promise((r) => setTimeout(r, 500));
    }

    // We transitioned to the new page. We need to get the interactions to do
    // on that page

    const values = {
      project: task.project,
      site: task.site,
      page,
      crawl_id: task.crawl_id,
    };
    const resp = await sendToServer("/sub_page_interactions", values);
    if (!resp) {
      await startNextTask();
      return;
    }

    task.page = page;
    task.interactions = JSON.parse(resp);
    chrome.scripting.executeScript({
      target: { tabId: sender.tab.id },
      func: doInteractions,
      args: [task],
    });
  }
});

async function startNextTask() {
  while (true) {
    const resp = await sendToServer("/next_task");
    if (!resp) {
      isCrawling = false;
      return;
    }

    const json = await JSON.parse(resp);

    if (!json?.status) {
      console.error("Could not retrieve next task");
      isCrawling = false;
      return;
    }

    if (json.status === "wait") {
      await new Promise((r) => setTimeout(r, 3000));
      continue;
    } else if (json.status !== "ready") {
      console.error(`Unknown crawling status: ${  json.status}`);
      isCrawling = false;
      return;
    }

    const task = json.task;
    if (!task || !task.url) {
      console.error("Invalid task");
      isCrawling = false;
      return;
    }

    const tab = await chrome.tabs.create({ url: task.url });
    if (!tab.id) {
      return;
    }
    while ((await chrome.tabs.get(tab.id)).status !== "complete") {
      await new Promise((r) => setTimeout(r, 500));
    }

    chrome.scripting.executeScript({
      target: { tabId: tab.id },
      func: doInteractions,
      args: [task],
    });
    break;
  }
}

/**
 * @param {any} task
 */
async function doInteractions(task) {
  async function completeTask() {
    await sendPage();
    chrome.runtime.sendMessage({ action: "taskCompleted" });
  }

  async function sendPage() {
    await chrome.runtime.sendMessage({
      action: "taskFinished",
      task,
      result: document.documentElement.outerHTML,
    });
  }

  const interactions = task.interactions;

  if (interactions.length === 0) {
    console.log("no interactions");
    await completeTask();
    return;
  }

  for (let i = 0; i < interactions.length; i++) {
    await new Promise((resolve) => setTimeout(resolve, 500));

    const interaction = interactions[i];

    if (interaction.action === "scroll_to_end") {
      let previousHeight = 0;
      while (true) {
        window.scrollTo(0, document.body.scrollHeight);
        const currentHeight = document.body.scrollHeight;
        if (currentHeight === previousHeight) {
          break;
        }
        await new Promise((resolve) => setTimeout(resolve, 500));
        previousHeight = currentHeight;
      }
    }

    if (interaction.action === "send_current") {
      await sendPage();
    }

    if (interaction.action === "click") {
      const xpath = interaction.annotation;
      const element = document.evaluate(
        xpath,
        document,
        null,
        XPathResult.FIRST_ORDERED_NODE_TYPE,
        null,
      ).singleNodeValue;
      if (!(element instanceof HTMLElement)) {
        continue;
      }
      if (interaction.navigates) {
        await sendPage();
        await chrome.runtime.sendMessage({
          action: "expectNavigation",
          task,
          page: interaction.navigates,
        });
      }

      await new Promise((resolve) => setTimeout(resolve, 100));
      element.click();
      if (interaction.navigates) {
        return;
      }
    }

    if (interaction.action === "input") {
      const xpath = interaction.annotation;
      const element = document.evaluate(
        xpath,
        document,
        null,
        XPathResult.FIRST_ORDERED_NODE_TYPE,
        null,
      ).singleNodeValue;
      if (!(element instanceof HTMLElement)) {
        continue;
      }
      element.focus();
      await new Promise((resolve) => setTimeout(resolve, 1000));

      const mouseOverEvent = new MouseEvent("mouseover", {
        view: window,
        bubbles: true,
        cancelable: true,
      });
      element.dispatchEvent(mouseOverEvent);

      const mouseClickEvent = new MouseEvent("click", {
        view: window,
        bubbles: true,
        cancelable: true,
      });
      element.dispatchEvent(mouseClickEvent);

      const inputString = interaction.data;
      if (element instanceof HTMLInputElement) {
        element.value = inputString;
      }
      const inputEvent = new InputEvent("input", {
        data: inputString,
        bubbles: true,
        cancelable: false,
      });
      element.dispatchEvent(inputEvent);
      await new Promise((resolve) => setTimeout(resolve, 1000));
    }
  }
  console.log("Finished interactions");

  await completeTask();
}

/**
 * @param {string} path - The realtive path on the server to send the data to
 * @param {object|null} postArgs - The values to send in the body of the post
 * @returns {Promise<string | null>} Either the response from the server, or null, if anything went wrong
 */
async function sendToServer(path, postArgs = null) {
  const { serverUrl, authToken } = await chrome.storage.local.get([
    "serverUrl",
    "authToken",
  ]);
  if (typeof serverUrl !== "string" || typeof authToken !== "string") {
    return null;
  }
  /**@type{any} */
  const req = {
    method: "GET",
    headers: {
      Authorization: `Bearer ${  authToken}`,
    },
  };
  if (postArgs) {
    req.method = "POST";
    req.headers["Content-Type"] = "application/json";
    req.body = JSON.stringify(postArgs);
  }

  const fullUrl = serverUrl + path;
  const response = await fetch(fullUrl, req);
  const serverText = await response.text();

  if (!response.ok) {
    const errorDesc = `Could not send request to '${fullUrl}' with body: ${JSON.stringify(postArgs)}. Server responded with: ${serverText}`;
    console.error(errorDesc);
    return null;
  }
  return serverText;
}
