/**
 * Does what document.getElementById does, but throws err on not found
 * @param {string} id - The id of the element to get
 * @returns {HTMLElement}
 */
function getElement(id) {
  const el = document.getElementById(id);
  if (el === null) {
    throw new Error(`get element by id '${id}' failed`);
  }
  return el;
}

/**
 * Gets a select element with the provided id and throws an error, if that is
 * not possible
 * @param {string} id - The id of the element to get
 * @returns {HTMLSelectElement | HTMLInputElement}
 */
function getInput(id) {
  const el = document.getElementById(id);
  if (el instanceof HTMLSelectElement || el instanceof HTMLInputElement) {
    return el;
  }
  throw new Error(`get element by id '${id}' failed`);
}

const annotationContainer = getElement("annotationContainer");
const siteContainer = getElement("siteContainer");
const pageContainer = getElement("pageContainer");
const projectContainer = getElement("projectContainer");
const startAnnotation = getElement("startAnnotation");

const startCrawling = getElement("startCrawling");
const stopCrawling = getElement("stopCrawling");
const overviewButton = getElement("overviewButton");
const setupContainer = getElement("setupContainer");
const loginButton = getElement("loginButton");
const logoutButton = getElement("logoutButton");

logoutButton.addEventListener("click", () => {
  chrome.storage.local.remove(["serverUrl", "authToken"], () => {
    window.location.reload();
  });
});

/** @type {Record<string, Project>|null} */
let config = null;

/** @type {string|null} */
let selectedProject = null;
/** @type {string|null} */
let selectedSite = null;
/** @type {string|null} */
let selectedPage = null;

function updateSelection() {
  hideElements(
    siteContainer,
    pageContainer,
    projectContainer,
    overviewButton,
    startAnnotation,
    startCrawling,
    stopCrawling,
    logoutButton,
  );

  clearSelection("project");
  clearSelection("site");
  clearSelection("page");

  if (config === null) {
    return;
  }
  showElements(overviewButton, logoutButton);
  projectContainer.style.display = "flex";

  for (const projectName in config) {
    addToSelection("project", projectName);
    selectedProject = selectedProject || projectName;
    if (selectedProject !== projectName) {
      continue;
    }
    const project = config[projectName];
    siteContainer.style.display = "flex";
    for (const siteName in project.sites) {
      addToSelection("site", siteName);
      selectedSite = selectedSite || siteName;
      if (selectedSite !== siteName) {
        continue;
      }
      pageContainer.style.display = "flex";
      for (const pageName in project.sites[siteName]) {
        addToSelection("page", pageName);
        selectedPage = selectedPage || pageName;
        startAnnotation.style.display = "grid";
        startCrawling.style.display = "block";
        stopCrawling.style.display = "block";

        if (selectedPage !== pageName) {
          continue;
        }
      }
    }
  }
  getInput("pageSelection").value = selectedPage ?? "";
  getInput("siteSelection").value = selectedSite ?? "";
  getInput("projectSelection").value = selectedProject ?? "";
}
updateSelection();

/**
 * @typedef {"project" | "site" | "page"} StageType
 */

/**
 * @param {StageType} stage
 * @param {string} name
 */
async function addEntry(stage, name) {
  if (name === null) {
    return;
  }
  if (stage === "project") {
    const resp = await sendToServer("/add_project", { project: name });
    if (resp === null) {
      return;
    }
    await tryUpdateServerConfig();
    selectedProject = name;
    selectedSite = null;
    selectedPage = null;
  } else if (stage === "site") {
    const resp = await sendToServer("/add_site", {
      project: selectedProject,
      site: name,
    });
    if (resp === null) {
      return;
    }
    await tryUpdateServerConfig();
    selectedSite = name;
    selectedPage = null;
  } else if (stage === "page") {
    const values = {
      project: selectedProject,
      site: selectedSite,
      page: name,
    };
    const resp = await sendToServer("/add_page", values);
    if (resp === null) {
      return;
    }
    await tryUpdateServerConfig();
    selectedPage = name;
  }
  updateSelection();
}

/**
 * @param {StageType} screen
 */
function startUpdateSelectionListener(screen) {
  const addButton = getElement(`${screen}AddButton`);
  const addInput = getElement(`${screen}AddInput`);
  if (!(addInput instanceof HTMLInputElement)) {
    return;
  }
  addButton.addEventListener("click", async () => {
    await addEntry(screen, addInput.value);
  });
}
startUpdateSelectionListener("project");
startUpdateSelectionListener("site");
startUpdateSelectionListener("page");

chrome.storage.local.get(
  ["serverUrl", "authToken"],
  async ({ serverUrl, authToken }) => {
    if (typeof serverUrl === "string" && typeof authToken === "string") {
      const res = await tryUpdateServerConfig();
      if (res) {
        finishSetup();
      }
    }
  },
);

/**
 * Tries to login to the server. If successfull, this will transition to the
 * project select and store the used server url
 * @param {string} url
 * @param {string} username
 * @param {string} password
 */
async function login(url, username, password) {
  /**
   * Hashes a string using sha256
   * @param {string} message
   * @returns {Promise<string>}
   */
  async function sha256(message) {
    const encoder = new TextEncoder();
    const data = encoder.encode(message);
    const hashBuffer = await crypto.subtle.digest("SHA-256", data);
    const hashArray = Array.from(new Uint8Array(hashBuffer));
    return hashArray.map((byte) => byte.toString(16).padStart(2, "0")).join("");
  }
  const pw_hash = await sha256(password);

  const response = await fetch(`${url}/login`, {
    method: "POST",
    headers: {
      "Content-Type": "application/json",
    },
    body: JSON.stringify({ username, pw_hash }),
  });

  if (!response.ok) {
    alert(`Login failed: ${await response.text()}`);
    return;
  }

  const authToken = await response.text();

  await chrome.storage.local.set({ serverUrl: url, authToken });
  if (!(await tryUpdateServerConfig())) {
    // We were not able to actually use these settings??
    return;
  }
  finishSetup();
}

function finishSetup() {
  annotationContainer.style.display = "grid";
  hideElements(setupContainer, startAnnotation);
  updateSelection();
}

/**
 * @return {Promise<boolean>}
 */
async function tryUpdateServerConfig() {
  const resp = await sendToServer("/list");
  if (resp === null) {
    return false;
  }
  const json = JSON.parse(resp);
  if (json) {
    config = json;
    return true;
  } else {
    return false;
  }
}

/**
 *
 * @param {string} relativeUrl
 * @param {object|null} postArgs
 * @returns {Promise<string|null>}
 */
async function sendToServer(relativeUrl, postArgs = null) {
  const storage = await chrome.storage.local.get(["serverUrl", "authToken"]);
  if (
    typeof storage.serverUrl !== "string" ||
    typeof storage.authToken !== "string"
  ) {
    return null;
  }

  /**
   * @type {{ method: "GET" | "POST"; headers: Record<string, string>; body?: string }}
   */
  const req = {
    method: "GET",
    headers: {
      Authorization: `Bearer ${storage.authToken}`,
    },
  };
  if (postArgs) {
    req.method = "POST";
    req.headers["Content-Type"] = "application/json";
    req.body = JSON.stringify(postArgs);
  }
  const fullUrl = storage.serverUrl + relativeUrl;
  const response = await fetch(fullUrl, req);
  const serverText = await response.text();
  if (!response.ok) {
    if (relativeUrl !== "/list") {
      const errorDesc = `Could not send request to '${fullUrl}' with body: ${JSON.stringify(postArgs)}. Server responded with: ${serverText}`;
      alert(errorDesc);
    }
    return null;
  }
  return serverText;
}

/**
 * @param {StageType} screen
 */
function startAddSelectionListener(screen) {
  const selection = getInput(`${screen}Selection`);

  selection.addEventListener("change", () => {
    if (screen === "project") {
      selectedProject = selection.value;
      selectedSite = null;
      selectedPage = null;
    } else if (screen === "site") {
      selectedSite = selection.value;
      selectedPage = null;
    } else if (screen === "page") {
      selectedPage = selection.value;
    }
    updateSelection();
  });
}
startAddSelectionListener("project");
startAddSelectionListener("site");
startAddSelectionListener("page");

loginButton.addEventListener("click", async () => {
  const inputServerUrl = getInput("setUrlInput").value;
  const inputPassword = getInput("setPasswordInput").value;
  const inputUsername = getInput("setUsernameInput").value;

  if (inputPassword && inputServerUrl && inputUsername) {
    await login(inputServerUrl, inputUsername, inputPassword);
  }
});

overviewButton.addEventListener("click", async () => {
  const { serverUrl } = await chrome.storage.local.get(["serverUrl"]);
  await chrome.tabs.create({ url: `${serverUrl}/overview` });
});

startCrawling.addEventListener("click", async () => {
  const endpoint = `/crawl?project=${selectedProject}&site=${selectedSite}`;
  const resp = await sendToServer(endpoint);
  if (resp === null) {
    return;
  }
  chrome.runtime.sendMessage({ action: "startCrawling" });
});

stopCrawling.addEventListener("click", async () => {
  await chrome.runtime.sendMessage({ action: "stopCrawling" });
});

/**
 * Validates the provided identifier.
 * An identifier is considered valid if it is at least 3 characters long
 * and contains only alphanumeric characters or underscores.
 *
 * @param {string} ident - The identifier to validate.
 * @returns {boolean} - Returns true if the identifier is valid, otherwise false.
 */
function isValidIdent(ident) {
  const isValidLength = ident.length >= 3;
  const isValidChars = /^[a-zA-Z0-9_]+$/.test(ident);
  return isValidLength && isValidChars;
}

/**
 * Sets the display style of all elements given to none
 * @param  {...HTMLElement | null} elements
 */
function hideElements(...elements) {
  elements.forEach((element) => {
    if (element) {
      element.style.display = "none";
    }
  });
}
/**
 * Sets the display style of all elements given to initial
 * @param  {...HTMLElement|null} elements
 */
function showElements(...elements) {
  elements.forEach((element) => {
    if (element) {
      element.style.display = "initial";
    }
  });
}
/**
 * Adds a project to the ui
 * @param {StageType} screen - The selection screen you want to add something to
 * @param {string} name - The name of the new project you want to add to the ui
 */
function addToSelection(screen, name) {
  const listItem = document.createElement("option");
  listItem.textContent = name;
  listItem.value = name;
  getInput(`${screen}Selection`).appendChild(listItem);
}
/**
 * Clears all child elements from the appropriate screen
 * @param {StageType} screen - The selection screen you want to add something to */
function clearSelection(screen) {
  getElement(`${screen}Selection`).innerHTML = "";
}

/**
 * Creates an event listener, that waits for a user to press the "Add" option
 * on the coresponding selection screen
 * @param {StageType} screen - The screen, that the listener should be created for
 */
function startAddButtonEventListener(screen) {
  getElement(`${screen}AddButton`).addEventListener("click", () => {
    const newName = getInput(`${screen}AddInput`).value;
    // TODO: actually send this to the server to validate & store
    if (!isValidIdent(newName)) {
      return;
    }
    addToSelection(screen, newName);
  });
}
startAddButtonEventListener("project");
startAddButtonEventListener("site");
startAddButtonEventListener("page");

startAnnotation.addEventListener("click", () => {
  if (!selectedProject || !selectedSite || !selectedPage || !config) {
    return;
  }
  const args = {
    project: selectedProject,
    site: selectedSite,
    page: selectedPage,
    config,
  };

  chrome.tabs.query({ active: true, currentWindow: true }, (tabs) => {
    if (!tabs[0].id) {
      return;
    }
    chrome.scripting.removeCSS({
      target: { tabId: tabs[0].id },
      files: ["styles_floating.css"],
    });
    chrome.scripting.insertCSS({
      target: { tabId: tabs[0].id },
      files: ["styles_floating.css"],
    });
    chrome.scripting.executeScript({
      target: { tabId: tabs[0].id },
      func: createFloatingWindow,
      args: [args],
    });
    window.close();
  });
});

///////////////////////////////////////////////////////////////////////////////

/**
 * @typedef {object} Project
 * @property {Record<string, Table>} tables - Tables this project has
 * @property {Record<string, string[]>} knowledge - A map of category names to values, where the values are arrays of strings.
 * @property {Record<string, Record<string, PageInfo>>} sites - A map of site information, where the keys are strings and the values are of type `SiteInfo`.
 */

/**
 * @typedef {object} BaseTable
 * @property {string} type
 * @property {boolean} has_timestamps
 * @property {Record<string, TableIndex>} indexes
 * @property {Record<string, Column | DataType>} columns
 */

/**
 * @typedef {BaseTable & { type: "primary" }} PrimaryTable
 */

/**
 * @typedef {BaseTable & { type: "secondary", primary_table: string }} SecondaryTable
 */

/**
 * @typedef {BaseTable & { type: "many_to_many", table_a: string, table_b: string }} ManyToManyTable
 */

/**
 * @typedef {PrimaryTable | SecondaryTable | ManyToManyTable} Table
 */

/**
 * Represents a unique index for a table.
 *
 * @typedef {object} TableIndex
 * @property {boolean} is_unique - Whether the index should be created as unique.
 * @property {string[]} columns - The list of column names that this index spans.
 */

/** @typedef {("int" | "float" | "text" | "references")} DataType

/**
 * Represents a reference to another table as a data type.
 *
 * @typedef {object} References
 * @property {string} table - The name of the table that this column references.
 */

/**
 * Represents a column definition in a table.
 *
 * @typedef {object} Column
 * @property {DataType} type - The data type that the column conforms to.
 * @property {boolean} [track_updates=false] - Whether to track updates on this column.
 */

/**
 * The key type for the table, which determines how the table is identified.
 *
 * @typedef {object} TableKey
 * @property {string} Primary - An auto-incrementing primary key for the table.
 * @property {Secondary} Secondary - A secondary key referencing another table's primary key.
 * @property {ManyToMany} ManyToMany - A many-to-many relationship between two tables.
 */

/**
 * Represents a secondary key, which is a reference to another table's primary key.
 *
 * @typedef {object} Secondary
 * @property {string} primary_table - The name of the table whose primary key this table references.
 */

/**
 * Represents a many-to-many relationship between two tables.
 *
 * @typedef {object} ManyToMany
 * @property {string} table_a - The name of the first table in the relationship.
 * @property {string} table_b - The name of the second table in the relationship.
 */

/**
 * @typedef {object} PageInfo
 * @property {string | null} start - The starting URL, that this page starts at. Will be inserted at the start of a new crawling run.
 * @property {boolean} is_search - If this page does not yield any unseen links from another table, it will not insert the links found for this table. Pages with this flag set will also not be cached.
 * @property {Record<string, NodeInfo>} nodes - A map of node information, where the keys are strings and the values are of type `NodeInfo`.
 * @property {BrowserInteraction[]} interactions - A list of browser interactions associated with the page.
 * @property {Record<string, string|ScoredXpath[]>} xpaths - A map of the name of the annotation to the xpath to reach that annotated node.
 */

/**
 * @typedef {object} ScoredXpath
 * @property {string} xp
 * @property {number} r_score
 * @property {number} f_score
 */

/**
 * @typedef { { replace_all: string, replacement: string } | { replace_once: string, replacement: string } | { debug_print: string } } Cleanup
 */

/**
 * @typedef {object} InputInteraction
 * @property {"input"} action
 * @property {string} annotation
 * @property {string} data
 */

/**
 * @typedef {object} ClickInteraction
 * @property {"click"} action
 * @property {string} annotation
 * @property {string} navigates
 */

/**
 * @typedef {object} ScrollToEndInteraction
 * @property {"scroll_to_end"} action
 */

/**
 * @typedef {object} SendCurrentInteraction
 * @property {"send_current"} action
 */

/**
 * @typedef {InputInteraction | ClickInteraction | ScrollToEndInteraction | SendCurrentInteraction} BrowserInteraction
 */

/**
 * @typedef {object} NodeInfo
 * @property {NodeSplit | null} node_split - Optional information about how this node is split.
 * @property {NodeTable | null} table - A table associated with this node, if applicable.
 * @property {NodePosition | null} position - Information about how this node is positioned relative to others.
 * @property {Record<string, DataPoint>} points - A map of values to extract from the page.
 */

/**
 * @typedef {object} DataPoint
 * @property {string} annotation - The name used when annotating HTML page nodes.
 * @property {string | null} extractor - The suffix added to the annotation XPath when evaluating the XPath to the annotation node.
 * @property {Cleanup[]} cleanups - A list of cleanup operations applied to the extracted value(s).
 * @property {DataTarget} target - The final destination of the extracted data.
 * @property {DataSelection} selection - The method used to select data when multiple results are found.
 */

/**
 * @typedef {{ type: "insert", column: string }} DataTargetNodeTableInsert
 */

/**
 * @typedef {{ type: "insert_secondary", table: string, column: string }} DataTargetSecondaryTableInsert
 */

/**
 * @typedef {{ type: "insert_mtm", column: string }} DataTargetManyToManyInsert
 */

/**
 * @typedef {{ type: "page_url", page: string, table_hint?: TableHint }} DataTargetPageUrl
 */

/**
 * @typedef {{ type: "auxiliary"}} DataTargetAuxiliary
 */

/**
 * @typedef {object} TableHint
 * @property {string} table_name - The name of the table being referenced.
 * @property {string} url_column - The column that contains the URL.
 * @property {string | undefined} non_null_column - An optional column that must be non-null.
 */

/**
 * @typedef {DataTargetNodeTableInsert | DataTargetSecondaryTableInsert | DataTargetManyToManyInsert | DataTargetPageUrl | DataTargetAuxiliary} DataTarget
 */

/**
 * @typedef {{ type: "take_first" }} DataSelectionTakeFirst
 */

/**
 * @typedef {{ type: "take_first_matches", regex: string }} DataSelectionTakeFirstMatches
 */

/**
 * @typedef {{ type: "take_first_non_empty"}} DataSelectionTakeFirstNonEmpty
 */

/**
 * @typedef {{ type: "take_n", n: number}} DataSelectionTakeN
 */

/**
 * @typedef {{type: "expect_one"}} DataSelectionExpectOne
 */

/**
 * @typedef {DataSelectionTakeFirst | DataSelectionTakeFirstMatches | DataSelectionTakeFirstNonEmpty | DataSelectionTakeN | DataSelectionExpectOne} DataSelection
 */

/**
 * @typedef {object} NodeSplit
 * @property {string} regex - The regular expression used for splitting.
 * @property {string | null} extractor - Optional extractor for the split data.
 */

/**
 * @typedef {object} NodeTable
 * @property {string} name - The table associated with this node.
 * @property {NodeTableIdent[]} idents - Identifiers used to find the target row in the database.
 * @property {ManyToManyIdent | null} mtm - The many-to-many table associated with this node, if applicable.
 */

/**
 * @typedef {object} ManyToManyIdent
 * @property {string} table - The name of the many-to-many table.
 * @property {NodeTableIdent[]} extra_idents - Additional identifiers needed to distinguish entries.
 */

/**
  * @typedef {object} Point
  * @property {'point'} type - Discriminator for DataPoint
  * @property {string} point - The datapoint name to look up in the table
  * @property {string} column - The corresponding column

  * @typedef {object} ParentId
  * @property {'parent_id'} type - Discriminator for ParentId
  * @property {string} column - The column in the target to match the parent id

  * @typedef {object} Url
  * @property {'url'} type - Discriminator for Url
  * @property {string} column - The column containing the page URL

  * @typedef {object} Constant
  * @property {'constant'} type - Discriminator for Constant
  * @property {string} column - The column name
  * @property {string} value - The constant value

  * @typedef {Point | ParentId | Url | Constant} NodeTableIdent
  */

/**
 * @typedef {object} NodePosition
 * @property {string} annotation - The annotation that marks all nodes of this type.
 * @property {string | null} parent_node - The name of the parent node, if applicable.
 */

/**
 * Creates a floating window to annotate the page and to configure the project.
 * @param {{ project: string, site: string, page: string, config: {[key: string]: Project} }} wArgs - The selection details for the current annotation context.
 */
async function createFloatingWindow(wArgs) {
  const { serverUrl, authToken } = await chrome.storage.local.get([
    "serverUrl",
    "authToken",
  ]);
  if (typeof serverUrl !== "string" || typeof authToken !== "string") {
    return;
  }

  document.getElementById("webdbFloatingWindow")?.remove();

  const floatingWindow = document.createElement("div");
  floatingWindow.id = "webdbFloatingWindow";

  const title = document.createElement("h3");
  title.id = "webdbTitle";
  title.textContent = `${wArgs.project}/${wArgs.site}/${wArgs.page}`;
  floatingWindow.appendChild(title);

  const selectionContainer = document.createElement("div");
  selectionContainer.id = "webdbbSelectionContainer";
  floatingWindow.appendChild(selectionContainer);

  const selectionBar = document.createElement("div");
  selectionBar.id = "webdbSelectionBar";
  selectionContainer.appendChild(selectionBar);

  ["annotation", "table", "nodes", "page", "xpaths", "knowledge"].forEach(
    (name) => {
      const buttonText = name.charAt(0).toUpperCase() + name.slice(1);
      addButton(selectionBar, buttonText, "green", () => {
        moveToPage(name);
      });
    },
  );

  const problemsButton = addButton(selectionBar, "O", "green", () => {
    moveToPage("errors");
  });

  const currentPage = {
    name: "annotation",
    /** @type {Record<string, string| undefined>} */
    args: {},
    /** @type {{name: string, args: Record<string, string | undefined>}[]} */
    prev: [],
  };

  /**
   * Moves to the desired page with the provided arguments
   * @param {string|null} name - The name of the page to move to
   * @param {Record<string, string| undefined>} args - The arguments this page should initially have access to
   * @param {boolean} storePage - Wether or not this should
   */
  function moveToPage(name, args = {}, storePage = true) {
    if (name && name !== "") {
      if (storePage) {
        const old = {
          name: currentPage.name,
          args: currentPage.args,
        };
        currentPage.prev.push(old);
      }
      currentPage.name = name;
      currentPage.args = args;
    }

    selectionBar.childNodes.forEach((button) => {
      if (button instanceof HTMLElement) {
        const isActive =
          currentPage.name[0].toLowerCase() ===
          (button.textContent ?? "")[0].toLowerCase();
        button.classList.toggle("webdbBlue", !isActive);
        button.classList.toggle("webdbGreen", isActive);
      }
    });

    const problems = projectProblems();
    if (Object.values(problems).every((a) => a.every((a) => a[1] === "warn"))) {
      problemsButton.classList.add("webdbGreen");
      problemsButton.classList.remove("webdbRed");
      problemsButton.textContent = "?";
    } else {
      problemsButton.classList.add("webdbRed");
      problemsButton.classList.remove("webdbGreen");
      problemsButton.textContent = "!";
    }
    problemsButton.style.marginLeft = "auto";
    problemsButton.style.width = "12px";

    document.getElementById("webdbTabView")?.remove();

    const tabView = document.createElement("div");
    tabView.id = "webdbTabView";

    /** @type {Record<string, (tab: HTMLElement) => void>} */
    const possiblePages = {
      nodes: renderNodeView,
      nodesPointAdd: renderPointAdd,
      nodesAdd: renderNodeAdd,
      table: renderTableView,
      tableAdd: renderTableAdd,
      page: renderPageView,
      xpaths: renderXpathView,
      annotation: renderAnnotationView,
      knowledge: renderKnowledgeView,
      errors: renderErrorView,
    };

    const renderFunction = possiblePages[currentPage.name];
    if (renderFunction) {
      renderFunction(tabView);
    } else {
      alert(`Invalid page: ${currentPage.name}`);
    }

    selectionContainer.appendChild(tabView);
  }
  moveToPage(null, {}, false);

  /**
   * @param {string} path - The realtive path on the server to send the data to
   * @param {object} values - The values to send in the body of the post
   * @param {string | null} pageNavigate - The page to navigate to, if the request succeeds
   * @returns {Promise<string | null>} Either the response from the server, or null, if anything went wrong
   */
  async function sendToServer(path, values, pageNavigate = "") {
    const strValues = JSON.stringify(values);
    const fullUrl = serverUrl + path;
    const response = await fetch(fullUrl, {
      method: "POST",
      headers: {
        "Content-Type": "application/json",
        Authorization: `Bearer ${authToken}`,
      },
      body: strValues,
    });

    const serverText = await response.text();
    if (!response.ok) {
      // const errorDesc = `Could not send request to '${fullUrl}' with body: ${strValues}. Server responded with: ${serverText}`;
      const errorDesc = `Error: ${serverText}`;
      alert(errorDesc);
      return null;
    }

    const configResp = await fetch(`${serverUrl}/list`, {
      headers: {
        Authorization: `Bearer ${authToken}`,
      },
    });
    const configText = await configResp.text();

    if (!configResp.ok) {
      alert("Could not update config");
      return null;
    }
    const json = JSON.parse(configText);
    if (!json) {
      alert("Invalid update resp");
      return null;
    }
    wArgs.config = json;
    if (pageNavigate !== null) {
      moveToPage(pageNavigate);
    }

    return serverText ?? "";
  }

  /**
   * @param {HTMLElement|null} tab
   * @param {string | null} labelText
   * @param {boolean} defaultChecked
   * @param {string} helpText
   * @returns {{ checkbox: HTMLInputElement, label: HTMLDivElement | null, container: HTMLDivElement }} - Returns an object containing the created input and label elements.
   */
  function addCheckBox(tab, labelText, defaultChecked = false, helpText) {
    const id = randomId();
    const checkboxContainer = document.createElement("div");
    checkboxContainer.style.display = "flex";
    checkboxContainer.style.gap = "20px";

    const checkbox = document.createElement("input");
    checkbox.id = id;
    checkbox.type = "checkbox";
    checkbox.checked = defaultChecked;
    checkbox.classList.add("webdbCheckbox");

    const label = createLabel(id, labelText, helpText);

    if (label) {
      checkboxContainer.appendChild(label);
    }
    checkboxContainer.appendChild(checkbox);

    if (tab) {
      tab.appendChild(checkboxContainer);
    }
    return { checkbox, label, container: checkboxContainer };
  }

  /**
   * Checks if the provided ident is correct
   * @param {string} rawName - The ident name you want to check
   * @returns {boolean}
   */
  function isValidIdent(rawName) {
    const name = rawName.trim();
    if (typeof name !== "string") {
      return false;
    }

    const isAlphanumericOrUnderscore = /^[a-zA-Z][a-zA-Z0-9_]*$/;

    return (
      isAlphanumericOrUnderscore.test(name) &&
      name.length > 2 &&
      name.length < 50 &&
      !["fetch_time", "update_time"].includes(name) &&
      !name.endsWith("_id")
    );
  }

  /**
   * Renders the point addition view in the floating window.
   * @param {HTMLElement|null} tab - The parent element to which the input will be added
   * @param {string|null} labelText - The label for the input element.
   * @param {string|null} defaultValue - The default value for the input
   * @param {string|null} placeholder - The default placeholder
   * @param {string|null} helpText - The default placeholder
   * @returns {{ input: HTMLInputElement, label: HTMLDivElement | null }} - Returns an object containing the created input and label elements.
   */
  function addInput(
    tab,
    labelText,
    defaultValue,
    placeholder = null,
    helpText = null,
  ) {
    const id = randomId();
    const label = createLabel(id, labelText, helpText);

    const input = document.createElement("input");
    input.id = id;
    input.type = "text";
    input.classList.add("webdbInput");
    input.value = defaultValue ?? "";
    input.placeholder = placeholder ?? labelText ?? defaultValue ?? "";

    if (tab) {
      if (label) {
        tab.appendChild(label);
      }
      tab.appendChild(input);
    }
    return { input, label };
  }

  function randomId() {
    return `webdb_${Math.random().toString(36).substring(2, 15)}`;
  }

  /**
   * Adds a selection option for a specific target.
   * @param {HTMLElement|null} tab - The parent element to which the selection will be added.
   * @param {string|null} labelText - The label for the select element.
   * @param {Array<string>|null} options - The options to be added to the select element.
   * @param {string|null} helpText - The optional help text
   * @returns {{ select: HTMLSelectElement, label: HTMLDivElement | null }} - Returns an object containing the created select and label elements.
   */
  function addSelect(tab, labelText = null, options = null, helpText = null) {
    const id = randomId();
    const label = createLabel(id, labelText, helpText);
    const select = document.createElement("select");
    select.id = id;
    select.name = id;
    select.classList.add("webdbSelect");

    addSelectOptions(select, options);

    if (tab !== null) {
      if (label) {
        tab.appendChild(label);
      }
      tab.appendChild(select);
    }

    return { select, label };
  }

  /**
   * Adds a button with default styles
   * @param {HTMLElement|null} tab - The parent element to which the selection will be added.
   * @param {string} buttonText - The text on the button
   * @param {"green" | "blue" | "red"} color - The color for the button
   * @param {EventListenerOrEventListenerObject} onClick - The thing the button does on a click
   * @returns {HTMLButtonElement} - Returns an object containing the created select and label elements.
   */
  function addButton(tab, buttonText, color, onClick) {
    const button = document.createElement("button");
    button.textContent = buttonText;
    let buttonColor;
    if (color === "green") {
      buttonColor = "webdbGreen";
    } else if (color === "blue") {
      buttonColor = "webdbBlue";
    } else if (color === "red") {
      buttonColor = "webdbRed";
    } else {
      throw new Error("Invalid color");
    }
    button.classList.add("webdbButton", buttonColor);
    if (onClick) {
      button.addEventListener("click", onClick);
    }
    tab?.appendChild(button);
    return button;
  }

  /**
   * @param {HTMLElement|null} tab
   * @param {string|null} id
   * @returns {HTMLDivElement}
   */
  function addTable(tab, id = null) {
    const tableDiv = document.createElement("div");
    tableDiv.id = id ?? randomId();
    tableDiv.classList.add("webdbTable");
    tab?.appendChild(tableDiv);
    return tableDiv;
  }

  /**
   * @param {HTMLElement|null} tab
   * @param {boolean} isTitle
   * @returns {HTMLDivElement}
   */
  function addRow(tab, isTitle = false) {
    const row = document.createElement("div");
    row.classList.add("webdbRow");
    if (isTitle === true) {
      row.classList.add("webdbRowTitle");
    }
    tab?.appendChild(row);
    return row;
  }

  /**
   * @param {HTMLElement|null} tab
   * @param {string|HTMLElement|null} element
   * @returns {HTMLElement}
   */
  function addColumn(tab, element) {
    if (typeof element === "string") {
      const p = document.createElement("div");
      p.textContent = element;
      element = p;
    } else if (element === null) {
      element = document.createElement("div");
    }
    element.classList.add("webdbColumn");
    tab?.appendChild(element);
    return element;
  }

  /**
   * Creates a label for element (id)
   * @param {string} id - The id of the thing to label
   * @param {string | null} labelText - The text the labelshould show
   * @param {string|null} helpText - The text to show on hover
   * @returns {HTMLDivElement | null}
   */
  function createLabel(id, labelText, helpText) {
    if (!labelText) {
      return null;
    }
    const labelContainer = document.createElement("div");
    labelContainer.style.display = "flex";
    labelContainer.style.alignItems = "center";
    labelContainer.style.gap = "10px";
    const label = document.createElement("label");
    label.textContent = labelText;
    label.htmlFor = id;
    label.title = helpText ?? "";
    labelContainer.appendChild(label);

    if (helpText) {
      const questionMark = document.createElement("span");
      questionMark.textContent = "?";
      questionMark.style.display = "flex";
      questionMark.style.alignItems = "center";
      questionMark.style.justifyContent = "center";
      questionMark.style.border = "1px solid gray";
      questionMark.style.borderRadius = "50%";
      questionMark.style.background = "whiteSmoke";
      questionMark.style.width = "14px";
      questionMark.style.height = "14px";
      questionMark.style.cursor = "pointer";
      questionMark.title = helpText;
      labelContainer.appendChild(questionMark);
    }
    return labelContainer;
  }

  /**
   * Adds options to the select
   *
   * @param {HTMLSelectElement} element - The select Element to add to.
   * @param {Array<string>|null} options - The options to be added to the select element.
   */
  function addSelectOptions(element, options) {
    if (options !== null) {
      const oldSelected = element.value;
      element.innerHTML = "";
      options.forEach((option) => {
        const targetOption = document.createElement("option");
        targetOption.textContent = capitalizeWords(option);
        targetOption.value = option;
        element.appendChild(targetOption);
      });
      if (options.includes(oldSelected)) {
        element.value = oldSelected;
      }
    }
  }

  /**
   * Capitalizes all words in a string
   * @param {string} str - the uncapitalized input
   * @returns {string} - The capitalized string
   */
  function capitalizeWords(str) {
    return str
      .split(/[_\s]+/)
      .map((word) => word.charAt(0).toUpperCase() + word.slice(1).toLowerCase())
      .join(" ");
  }

  /**
   * Sets the display style of all elements given to none
   * @param  {...HTMLElement | null} elements
   */
  function hideElements(...elements) {
    elements.forEach((element) => {
      if (element) {
        element.style.display = "none";
      }
    });
  }
  /**
   * Sets the display style of all elements given to initial
   * @param  {...HTMLElement|null} elements
   */
  function showElements(...elements) {
    elements.forEach((element) => {
      if (element?.style.gap === "10px") {
        element.style.display = "flex";
      } else if (element) {
        element.style.display = "initial";
      }
    });
  }

  /**
   * @param {HTMLElement} tabView
   */
  function renderPointAdd(tabView) {
    // These two may be null/undefined
    const node = currentPage.args.node;

    // Info about all nodes/columns/pages to be used in selects
    const projectConfig = wArgs.config[wArgs.project] ?? {};
    const nodes = projectConfig.sites[wArgs.site]?.[wArgs.page]?.nodes ?? {};
    const nodeNames = Object.keys(nodes);
    const site = projectConfig.sites[wArgs.site] ?? {};
    const pages = Object.keys(site);
    const oldPointName = currentPage.args.point;
    const oldPoint =
      oldPointName && node ? nodes[node]?.points[oldPointName] : null;
    const annotation = oldPoint?.annotation ?? currentPage.args.annotation;

    const annotations = getAllAnnotations();
    if (annotation && !annotations.includes(annotation)) {
      annotations.push(annotation);
    }
    const { select: annotationSelect } = addSelect(
      tabView,
      "Annotation",
      annotations,
      "Points get their data from an html node. The annotation denotes which nodes to get the data from",
    );
    if (annotation) {
      annotationSelect.value = annotation;
    }

    const { input: pointNameInput } = addInput(
      tabView,
      "Point Name",
      oldPointName ?? annotation ?? annotationSelect.value,
    );

    if (oldPointName) {
      pointNameInput.disabled = true;
    }

    annotationSelect.addEventListener("change", () => {
      if (
        annotations.includes(pointNameInput.value) ||
        pointNameInput.value === ""
      ) {
        pointNameInput.value = annotationSelect.value;
      }
    });

    const { select: extractorSelect } = addSelect(
      tabView,
      "Extractor",
      null,
      "This chooses what to select from the html element. This is most commonly just the text, but may need to become @href to extract a link, or other attributes",
    );
    /**
     * @param {boolean} isFirst
     */
    function update_extractors(isFirst = false) {
      const validExtractors = ["string()", "text()"];
      if (!annotationSelect.value) {
        addSelectOptions(extractorSelect, validExtractors);
        return;
      }
      const annotatedNodes = getNodesForAnnotation(annotationSelect.value);
      for (const annotatedNode of annotatedNodes) {
        for (const attr of annotatedNode.attributes) {
          const atrKey = `@${attr.name}`;
          if (!validExtractors.includes(atrKey) && atrKey !== "@webdb") {
            validExtractors.push(atrKey);
          }
        }
      }
      addSelectOptions(extractorSelect, validExtractors);
      if (isFirst && oldPoint?.extractor) {
        if (!validExtractors.includes(oldPoint.extractor)) {
          validExtractors.push(oldPoint.extractor);
          addSelectOptions(extractorSelect, validExtractors);
        }
        extractorSelect.value = oldPoint?.extractor;
      }
    }
    update_extractors(true);
    annotationSelect.addEventListener("change", () => update_extractors());

    const label = createLabel(
      "webdbPreviewLabel",
      "Preview",
      "This is a preview of what the extracted data would look like",
    );
    if (label) {
      tabView.appendChild(label);
    }

    const cleanups = oldPoint?.cleanups ?? [];

    const cleanupTable = addTable(tabView);
    const preview = document.createElement("div");
    preview.id = "webdbPreview";
    tabView.appendChild(preview);

    function renderCleanups() {
      cleanupTable.innerHTML = "";
      const infoRow = addRow(cleanupTable, true);
      addColumn(infoRow, "Cleanup Type");
      for (const cleanup of cleanups) {
        const row = addRow(cleanupTable);
        if ("replace_all" in cleanup) {
          addColumn(row, "Replace All");
          addColumn(row, cleanup.replace_all);
          addColumn(row, cleanup.replacement);
          const rmButton = addButton(null, "X", "red", () => {
            const pos = cleanups.indexOf(cleanup);
            if (pos !== -1) {
              cleanups.splice(pos, 1);
            }
            renderCleanups();
          });
          rmButton.style.width = "12px";
          rmButton.style.height = "12px";
          rmButton.style.flex = "initial";
          rmButton.style.verticalAlign = "middle";
          addColumn(row, rmButton);
        } else if ("replace_once" in cleanup) {
          addColumn(row, "Replace Once");
          addColumn(row, cleanup.replace_once);
          addColumn(row, cleanup.replacement);
          const rmButton = addButton(null, "X", "red", () => {
            const pos = cleanups.indexOf(cleanup);
            if (pos !== -1) {
              cleanups.splice(pos, 1);
            }
            renderCleanups();
          });
          rmButton.style.width = "12px";
          rmButton.style.height = "12px";
          rmButton.style.flex = "initial";
          rmButton.style.verticalAlign = "middle";
          addColumn(row, rmButton);
        }
      }
      const editRow = addRow(cleanupTable);
      const typeSelect = addSelect(null, null, [
        "Replace Once",
        "Replace All",
      ]).select;
      addColumn(editRow, typeSelect);
      const replace = addInput(null, null, null).input;
      addColumn(editRow, replace);
      const replacement = addInput(null, null, null).input;
      addColumn(editRow, replacement);
      const add = addButton(null, "Add", "blue", () => {
        if (typeSelect.value === "Replace All") {
          cleanups.push({
            replace_all: replace.value,
            replacement: replacement.value,
          });
        } else if (typeSelect.value === "Replace Once") {
          cleanups.push({
            replace_once: replace.value,
            replacement: replacement.value,
          });
        }
        renderCleanups();
      });
      addColumn(editRow, add);
      updatePreview();
    }
    renderCleanups();

    function updatePreview() {
      const elem = getNodesForAnnotation(annotationSelect.value)[0];
      let previewText = "";
      if (elem === undefined) {
        // Default value
      } else if (extractorSelect.value === "string()") {
        previewText = elem.textContent ?? "";
      } else if (extractorSelect.value === "text()") {
        const text = elem.childNodes
          .values()
          .find((a) => a.nodeType === a.TEXT_NODE);
        previewText = text?.textContent ?? "";
      } else if (extractorSelect.value === "@class") {
        const attr = Array.from(elem.classList.values());
        previewText = attr.join(" ");
      } else if (extractorSelect.value.startsWith("@")) {
        const text = extractorSelect.value.substring(1);
        const attr = elem.getAttribute(text);
        previewText = attr ?? "";
      }

      for (const cleanup of cleanups) {
        if ('replace_all' in cleanup) {
          const re = new RegExp(cleanup.replace_all, 'gu');
          previewText = previewText.replaceAll(re, cleanup.replacement);
        } else if ('replace_once' in cleanup) {
          const re = new RegExp(cleanup.replace_once, 'u');
          previewText = previewText.replace(re, cleanup.replacement);
        }
      }

      preview.textContent = previewText;
    }
    updatePreview();
    annotationSelect.addEventListener("change", updatePreview);
    extractorSelect.addEventListener("change", updatePreview);

    const { select: nodeSelect } = addSelect(tabView, "Node", nodeNames);
    if (node && nodeNames.includes(node)) {
      nodeSelect.value = node;
    }

    const tableInsertOptions = [
      "Insert Table",
      "Insert Secondary Table",
      "Insert MtM Table",
      "Page Link",
      "-- None --",
    ];
    const { select: targetSelect } = addSelect(
      tabView,
      "Data Target",
      tableInsertOptions,
    );
    const { select: columnSelect, label: columnSelectLabel } = addSelect(
      tabView,
      "Insertion Column",
    );

    let columnAdd = document.createElement("div");
    tabView.appendChild(columnAdd);

    const { select: pageSelect, label: pageSelectLabel } = addSelect(
      tabView,
      "Page Link:",
      pages,
    );

    function updateColumns() {
      hideElements(
        columnAdd,
        columnSelect,
        columnSelectLabel,
        pageSelect,
        pageSelectLabel,
      );

      /**
       * @param {boolean} isMtm
       */
      function resetTableColumns(isMtm) {
        showElements(columnSelect, columnSelectLabel);
        const node =
          projectConfig.sites[wArgs.site]?.[wArgs.page]?.nodes?.[
            nodeSelect.value
          ] || {};

        const tableName = isMtm ? node.table?.mtm?.table : node.table?.name;
        if (tableName) {
          showElements(columnAdd);
          const columnNames = Object.keys(
            projectConfig.tables?.[tableName]?.columns ?? {},
          );
          addSelectOptions(columnSelect, columnNames);
          const newNode = renderColumnCreation(null, tableName);
          columnAdd.replaceWith(newNode);
          columnAdd = newNode;
        } else {
          columnSelect.innerHTML = "";
          const noValidColumnOption = document.createElement("option");
          noValidColumnOption.value = "";
          noValidColumnOption.textContent = "Node does not have a table";
          columnSelect.appendChild(noValidColumnOption);
        }
      }

      if (targetSelect.value === "Insert Table") {
        resetTableColumns(false);
      } else if (targetSelect.value === "Insert MtM Table") {
        resetTableColumns(true);
      } else if (targetSelect.value === "Insert Secondary Table") {
        //
      } else if (targetSelect.value === "Page Link") {
        showElements(pageSelectLabel, pageSelect);
      }
    }

    nodeSelect.addEventListener("change", updateColumns);
    targetSelect.addEventListener("change", updateColumns);

    const oldTableName = node && nodes[node]?.table?.name;
    const oldTable = oldTableName
      ? projectConfig.tables[oldTableName]
      : undefined;
    if (oldPoint) {
      if (oldPoint.target.type === "insert") {
        targetSelect.value = "Insert Table";
        const oldColumn = oldTable?.columns[oldPoint.target.column];
        if (oldColumn) {
          columnSelect.value = oldPoint.target.column;
        }
      } else if (oldPoint.target.type === "insert_secondary") {
        // Handle insert_secondary case
        targetSelect.value = "Insert Secondary Table";
        const oldColumn = oldTable?.columns[oldPoint.target.column];
        if (oldColumn) {
          columnSelect.value = oldPoint.target.column;
        }
      } else if (oldPoint.target.type === "insert_mtm") {
        targetSelect.value = "Insert MtM Table";
        const oldMtMTableName = node && nodes[node]?.table?.mtm?.table;
        const oldMtMTable = oldMtMTableName
          ? projectConfig.tables[oldMtMTableName]
          : undefined;
        const oldColumn = oldMtMTable?.columns[oldPoint.target.column];
        if (oldColumn) {
          columnSelect.value = oldPoint.target.column;
        }
      } else if (oldPoint.target.type === "page_url") {
        targetSelect.value = "Page Link";
        if (site[oldPoint.target.page] !== undefined) {
          pageSelect.value = oldPoint.target.page;
        }
      } else if (oldPoint.target.type === "auxiliary") {
        targetSelect.value = "-- None --";
      }
    }
    updateColumns();

    const addPoint = async () => {
      const nodeName = nodeSelect.value.trim();
      const pointName = pointNameInput.value.trim();
      const extractor = extractorSelect.value.trim();
      const annotation = annotationSelect.value.trim();

      let target = null;
      if (targetSelect.value === "Insert Table") {
        target = {
          type: "insert",
          column: columnSelect.value,
        };
      } else if (targetSelect.value === "Page Link") {
        target = {
          type: "page_url",
          table: pageSelect.value,
        };
      } else if (targetSelect.value === "Insert MtM Table") {
        target = {
          type: "insert_mtm",
          column: columnSelect.value,
        };
      } else if (targetSelect.value === "-- None --") {
        target = {
          type: "auxiliary",
        };
      } else {
        alert("Invalid target");
      }

      const values = {
        project: wArgs.project,
        site: wArgs.site,
        page: wArgs.page,
        node: nodeName,
        data_point: pointName,
        annotation,
        extractor,
        target,
        cleanups,
        selection: {
          type: "take_first",
        },
        is_edit: Boolean(oldPoint),
      };
      await sendToServer("/add_point", values, "nodes");
    };
    const addPointButton = addButton(
      tabView,
      oldPoint ? "Save" : "Add Point",
      "green",
      addPoint,
    );

    function isValidInput() {
      const pointName = pointNameInput.value.trim();
      const nodeName = nodeSelect.value.trim();
      const annotation = annotationSelect.value.trim();

      if (
        nodeName === "" ||
        !isValidIdent(pointName) ||
        !isValidIdent(annotation) ||
        targetSelect.value === "Insert Secondary Table"
      ) {
        return false;
      }

      // TODO: Table Inserts
      return true;
    }
    startInputValidation(
      isValidInput,
      addPointButton,
      pointNameInput,
      annotationSelect,
      nodeSelect,
    );
  }

  /**
   * @param {HTMLElement} tabView
   */
  function renderNodeAdd(tabView) {
    const projectConfig = wArgs.config[wArgs.project] ?? {};
    const nodes = projectConfig.sites[wArgs.site]?.[wArgs.page]?.nodes ?? {};

    const oldNodeName = currentPage.args.node;
    const oldNodeInfo = oldNodeName && nodes ? nodes[oldNodeName] : null;

    const { input: nodeNameInput } = addInput(
      tabView,
      "Node Name",
      oldNodeName ?? currentPage.args.annotation ?? "",
    );
    if (oldNodeName) {
      nodeNameInput.disabled = true;
    }

    const tables = projectConfig.tables;
    const tableNames = Object.keys(tables);
    const { select: tableSelect } = addSelect(
      tabView,
      "Select Table",
      tableNames,
    );
    const oldTable = oldNodeInfo?.table?.name;
    if (tableNames.includes(oldTable ?? "")) {
      tableSelect.value = oldTable ?? "";
    }

    let idents = oldNodeInfo?.table?.idents ?? [];
    const identsTable = addTable(tabView);

    function renderIdents() {
      identsTable.innerHTML = "";
      const titleBar = addRow(identsTable, true);
      ["Point", "Column", "Data"].forEach((a) => {
        addColumn(titleBar, a);
      });

      for (const ident of idents) {
        const identRow = addRow(identsTable);
        addColumn(identRow, ident.type);
        addColumn(identRow, ident.column);

        const bonus =
          ident.type === "point"
            ? ident.point
            : ident.type === "constant"
              ? ident.value
              : "";
        addColumn(identRow, bonus);
        const rmButton = addButton(null, "X", "red", () => {
          const pos = idents.indexOf(ident);
          if (pos !== -1) {
            idents.splice(pos, 1);
          }
          renderIdents();
        });
        rmButton.style.width = "12px";
        rmButton.style.height = "12px";
        rmButton.style.flex = "initial";
        rmButton.style.verticalAlign = "middle";
        addColumn(identRow, rmButton);
      }

      const insertRow = addRow(identsTable);
      const { select: typeSelection } = addSelect(null, null, [
        "point",
        "parent_id",
        "url",
        "constant",
      ]);
      addColumn(insertRow, typeSelection);

      const selectedTable = tables[tableSelect.value];
      const columns = Object.keys(selectedTable?.columns ?? {});

      if (selectedTable) {
        columns.push(`${tableSelect.value}_id`);
      }

      const { select: columnSelection } = addSelect(null, null, columns);
      addColumn(insertRow, columnSelection);

      /**@type {HTMLElement} */
      let bonus = document.createElement("div");

      function updateBonus() {
        bonus.innerHTML = "";

        /**@type{HTMLElement} */
        let newBonus;
        if (typeSelection.value === "point") {
          const dataPoints = Object.keys(oldNodeInfo?.points ?? {});
          newBonus = addSelect(null, null, dataPoints).select;
        } else if (typeSelection.value === "constant") {
          newBonus = addInput(null, null, "").input;
        } else {
          newBonus = document.createElement("div");
          bonus.replaceWith(newBonus);
          bonus = newBonus;
          return;
        }
        const n = addColumn(null, newBonus);
        bonus.replaceWith(n);
        bonus = n;
      }
      addColumn(insertRow, bonus);
      typeSelection.addEventListener("change", updateBonus);
      updateBonus();
      const addIdentButton = addColumn(
        insertRow,
        addButton(null, "+", "blue", () => {
          const identType = typeSelection.value;
          if (identType === "point") {
            if (!(bonus instanceof HTMLSelectElement) || !bonus.value) {
              return;
            }
            idents.push({
              type: identType,
              column: columnSelection.value,
              point: bonus.value,
            });
          } else if (identType === "constant") {
            if (!(bonus instanceof HTMLInputElement) || !bonus.value) {
              return;
            }
            idents.push({
              type: identType,
              column: columnSelection.value,
              value: bonus.value,
            });
          } else if (identType === "parent_id" || identType === "url") {
            idents.push({
              type: identType,
              column: columnSelection.value,
            });
          } else {
            alert("Invalid Ident type");
          }

          renderIdents();
        }),
      );
      addIdentButton.style.width = "12px";
      addIdentButton.style.height = "12px";
      addIdentButton.style.flex = "initial";
      addIdentButton.style.verticalAlign = "middle";
    }

    renderIdents();
    tableSelect.addEventListener("change", () => {
      idents = [];
      renderIdents();
    });

    const oldAnnotation =
      currentPage.args.annotation ??
      oldNodeInfo?.position?.annotation ??
      (oldNodeInfo ? "-- ROOT --" : undefined);

    const annotations = getAllAnnotations();
    if (oldAnnotation && !annotations.includes(oldAnnotation)) {
      annotations.push(oldAnnotation);
    }
    annotations.push("-- ROOT --");

    const { select: annotationSelect } = addSelect(
      tabView,
      "Annotation",
      annotations,
      "The annotation, that defines where this node, so most likely list is going to have it's elements. If you do not select an annotation (''), this will we regarded as the root node",
    );

    const parents = Object.keys(nodes);
    const { select: parentSelect, label: parentLabel } = addSelect(
      tabView,
      "Parent Node",
      parents,
      "The node, that is above this in the HTML hierarchy. In almost most cases, this is going to be root, but if you have a list of roles per actor in a movie, you want to set this to actor here",
    );

    const { select: mtmTableSelect, label: mtmTableLabel } = addSelect(
      tabView,
      "Select MtM Table",
    );

    function noMtM() {
      addSelectOptions(mtmTableSelect, ["-- No MtM Relation --"]);
    }

    function checkParent() {
      hideElements(parentSelect, parentLabel, mtmTableSelect, mtmTableLabel);
      if (!annotationSelect.value || annotationSelect.value === "-- ROOT --") {
        noMtM();
        return;
      }
      showElements(parentSelect, parentLabel, mtmTableSelect, mtmTableLabel);
      /**@type{string | null} */
      const parentTableName = nodes[parentSelect.value]?.table?.name ?? "";
      const ourTableName = tableSelect.value;

      if (!parentTableName || ourTableName === parentTableName) {
        noMtM();
        return;
      }
      const options = ["-- None --"];
      for (const tableName of tableNames) {
        const table = tables[tableName];
        const type = table?.type;
        if (type !== "many_to_many") {
          continue;
        }
        const p = [ourTableName, parentTableName];
        if (
          p.includes(table.table_a ?? "") &&
          p.includes(table.table_b ?? "")
        ) {
          options.push(tableName);
        }
      }
      if (options.length === 1) {
        noMtM();
        return;
      }
      addSelectOptions(mtmTableSelect, options);
    }

    if (oldAnnotation) {
      annotationSelect.value = oldAnnotation;
    }
    const oldParent = oldNodeInfo?.position?.parent_node;
    if (oldParent) {
      parentSelect.value = oldParent;
    }

    annotationSelect.addEventListener("change", checkParent);
    parentSelect.addEventListener("change", checkParent);
    checkParent();

    const oldMtM = oldNodeInfo?.table?.mtm?.table;
    if (oldMtM) {
      mtmTableSelect.value = oldMtM;
    }

    const addNode = async () => {
      const newNodeName = nodeNameInput.value.trim();
      const annotation = annotationSelect.value.trim();
      const tableName = tableSelect.value.trim();
      const mtm = mtmTableSelect.value.trim();
      const parent = parentSelect.value.trim();

      if (!isValidIdent(newNodeName)) {
        return;
      }

      let position = null;
      if (annotation && annotation !== "" && annotation !== "-- ROOT --") {
        if (parent) {
          position = {
            annotation,
            parent_node: parent,
          };
        } else {
          position = { annotation };
        }
      }
      let table = null;
      if (mtm.startsWith("--")) {
        table = {
          name: tableName,
          idents,
        };
      } else {
        table = {
          name: tableName,
          idents,
          mtm: {
            table: mtm,
          },
        };
      }

      const values = {
        project: wArgs.project,
        site: wArgs.site,
        page: wArgs.page,
        node: newNodeName,
        position,
        table,
        is_edit: Boolean(oldNodeName),
      };
      await sendToServer("/add_node", values, "nodes");
    };

    /**
     *
     * @returns {boolean}
     */
    function isValid() {
      const newNodeName = nodeNameInput.value.trim();
      const table = tableSelect.value.trim();
      return isValidIdent(newNodeName) && table !== "";
    }

    const finishText = oldNodeName ? "Save" : "Add Node";
    const button = addButton(tabView, finishText, "green", addNode);

    startInputValidation(isValid, button, nodeNameInput, tableSelect);
  }

  /** @typedef {("warn" | "error")} LogLevel

  /**
   *
   * @returns {Record<string, [string, LogLevel, string|null][]>} A description of the error, as well as a
   *  you can press to go to the page, where you can fix this
   */
  function projectProblems() {
    /** @type {Record<string, [string, LogLevel, string|null][]>} */
    const res = {};
    const projectConfig = wArgs.config[wArgs.project];
    const pages = projectConfig.sites[wArgs.site];

    let hasStart = false;

    for (const [pageName, page] of Object.entries(pages)) {
      const isCurrentPage = pageName === wArgs.page;

      /**@type{[string, LogLevel, string|null][]} */
      const ourRes = (res[pageName] = []);

      if (page.start) {
        hasStart = true;
      }
      const xpathedAnnotations = new Set();
      for (const xpathName in page.xpaths) {
        xpathName
          .split(".")
          .forEach((part) => xpathedAnnotations.add(part.trim()));
      }

      for (const interaction of page.interactions) {
        if (interaction.action === "click" || interaction.action === "input") {
          if (xpathedAnnotations.has(interaction.annotation)) {
            ourRes.push([
              `An interaction is missing an Xpath for annotation '${interaction.annotation}', so it will never be extracted. Go to the Xpath tab and generate them`,
              "error",
              isCurrentPage ? "xpaths" : null,
            ]);
          }
        }
        if (interaction.action === "input") {
          const data = projectConfig.knowledge[interaction.data];
          if (data === undefined || data.length === 0) {
            ourRes.push([
              `An input interaction is refering to a knowledge key ('${interaction.data}'), that either does not exist, or that is empty. Define that knowledge in the knowledge tab`,
              "error",
              "knowledge",
            ]);
          }
        }
      }

      for (const [nodeName, node] of Object.entries(page.nodes)) {
        if (node.table?.name !== undefined) {
          const table = projectConfig.tables[node.table?.name];
          if (table === undefined) {
            ourRes.push([
              `Node ${nodeName} references table '${node.table?.name}', that does not exist`,
              "error",
              isCurrentPage ? "nodes" : null,
            ]);
            continue;
          }
          if (
            Object.values(node.table.idents).length === 0 &&
            Object.values(node.points).length > 0
          ) {
            ourRes.push([
              `Node ${nodeName} does not have any idents set, but contains at least one point. Edit the node and set at least one ident to make table insertions work`,
              "error",
              isCurrentPage ? "nodes" : null,
            ]);
            continue;
          }
          for (const ident of Object.values(node.table.idents)) {
            const column = table.columns[ident.column];
            if (column === undefined) {
              ourRes.push([
                `An ident for node ${nodeName} references a column, that does not exist`,
                "error",
                isCurrentPage ? "nodes" : null,
              ]);
              continue;
            }
            const columnType =
              typeof column === "string" ? column : column.type;
            if (ident.type === "url" || ident.type === "constant") {
              if (columnType !== "text") {
                ourRes.push([
                  `A text ident for node ${nodeName} references a column (${ident.column}), who's type is not text`,
                  "error",
                  isCurrentPage ? "nodes" : null,
                ]);
                continue;
              }
            } else if (ident.type === "parent_id") {
              if (columnType !== "int" && columnType !== "references") {
                ourRes.push([
                  `A parent_id ident for node ${nodeName} references a column (${ident.column}), who's type is not int`,
                  "error",
                  isCurrentPage ? "nodes" : null,
                ]);
                continue;
              }
            } else if (ident.type === "point") {
              const point = node.points[ident.point];
              if (point === undefined) {
                ourRes.push([
                  `A point ident for node ${nodeName} references a point (${ident.point}), that does not exist`,
                  "error",
                  isCurrentPage ? "nodes" : null,
                ]);
                continue;
              }
            }
          }
          const mtmTableName = node.table.mtm?.table;
          const mtmTable = mtmTableName
            ? projectConfig.tables[mtmTableName]
            : undefined;

          if (mtmTableName !== undefined) {
            if (mtmTable === undefined) {
              ourRes.push([
                `Node ${nodeName} references mtm table '${mtmTableName}', but that table does not exist`,
                "error",
                isCurrentPage ? "nodes" : null,
              ]);
              continue;
              // TODO: Check extra idents, but right now we can not even define
              // them, so..
            } else if (mtmTable?.type !== "many_to_many") {
              ourRes.push([
                `Node ${nodeName} references mtm table '${mtmTableName}', but that table is not a mtm table`,
                "error",
                isCurrentPage ? "nodes" : null,
              ]);
              continue;
            }
          }
        }
        if (node.position) {
          if (!xpathedAnnotations.has(node.position.annotation)) {
            ourRes.push([
              `The node '${nodeName}' does not have an xpath, associated with it, so none of its points will ever get evaluated. You must generate xpaths for it`,
              "error",
              isCurrentPage ? "xpaths" : null,
            ]);
          }
          if (
            node.position.parent_node &&
            page.nodes[node.position.parent_node] === undefined
          ) {
            ourRes.push([
              `The node '${nodeName}' is set to be below ${node.position.parent_node}, but that node does not exist`,
              "error",
              isCurrentPage ? "node" : null,
            ]);
          }
        }
        for (const [pointName, point] of Object.entries(node.points)) {
          if (!xpathedAnnotations.has(point.annotation)) {
            ourRes.push([
              `The point '${pointName}' on node '${nodeName}' does not have an xpath, associated with it, so it will never get extracted. You must generate xpaths for it`,
              "error",
              isCurrentPage ? "xpaths" : null,
            ]);
          }

          if (
            point.target.type === "page_url" &&
            pages[point.target.page] === undefined
          ) {
            ourRes.push([
              `The point '${pointName}' references a page, that does not exist. Either create that page, or change the point to point to a valid page`,
              "warn",
              isCurrentPage ? "node" : null,
            ]);
          } else if (
            point.target.type === "insert" &&
            (!node.table || projectConfig.tables[node.table.name] === undefined)
          ) {
            ourRes.push([
              `The point '${pointName}' is trying to do a table insert, but either the referenced column, or the entire table does not exist. Edit the point, or create the column`,
              "error",
              isCurrentPage ? "node" : null,
            ]);
          } else if (
            point.target.type === "insert_mtm" &&
            (!node.table ||
              !node.table.mtm ||
              projectConfig.tables[node.table.mtm.table] === undefined)
          ) {
            ourRes.push([
              `The point '${pointName}' is trying to do a mtm table insert, but either the referenced column, or the entire table does not exist. Edit the point, or create the column`,
              "error",
              isCurrentPage ? "node" : null,
            ]);
          }
        }
      }

      if (isCurrentPage) {
        for (const annotation of getAllAnnotations()) {
          if (!xpathedAnnotations.has(annotation)) {
            ourRes.push([
              `Annotation '${annotation}' is missing an Xpath, so it will never be extracted. Go to the Xpath tab and generate them`,
              "error",
              "xpaths",
            ]);
          }
        }
      }
    }

    /**@type{[string, LogLevel, string|null][]} */
    const globalRes = (res["global"] = []);

    if (!hasStart) {
      globalRes.push([
        "No starting URL has been set. This will lead to no pages getting crawled. Go to the 'page' tab and set a URL on the page you want to start crawling on",
        "error",
        "page",
      ]);
    }
    return res;
  }

  /**
   * @param {HTMLElement} tabView - The current tab
   */
  function renderErrorView(tabView) {
    const problems = projectProblems();

    const problemsTable = addTable(tabView);
    for (const [pageName, problemEntries] of Object.entries(problems)) {
      if (problemEntries.length === 0) {
        continue;
      }
      const pageRow = addRow(problemsTable, true);
      const t = addColumn(pageRow, `Problems for page ${pageName}`);
      t.style.flex = "4";

      addColumn(pageRow, "Level");
      addColumn(pageRow, "Resolve Tab");

      problemEntries.forEach(([errorDescription, logLevel, pageNavigate]) => {
        const problemRow = addRow(problemsTable);
        const errorColumn = addColumn(problemRow, errorDescription);
        errorColumn.style.flex = "4";
        const elem = document.createElement("p");
        elem.textContent = logLevel;
        addColumn(problemRow, elem);

        if (pageNavigate) {
          const b = addButton(null, "Fix", "blue", () => {
            moveToPage(pageNavigate);
          });
          b.style.height = "12px";
          addColumn(problemRow, b);
        } else {
          addColumn(problemRow, "");
        }
      });
    }
  }

  /**
   * @param {HTMLElement} tabView - The current tab
   */
  function renderKnowledgeView(tabView) {
    const projectConfig = wArgs.config[wArgs.project];
    const table = addTable(tabView);

    const knowledge = structuredClone(projectConfig.knowledge);

    function renderKnowledge() {
      table.innerHTML = "";
      const infoRow = addRow(table, true);
      addColumn(infoRow, "Name");
      const v = addColumn(infoRow, "Values");
      v.style.flex = "3";
      addColumn(infoRow, "");
      for (const [name, values] of Object.entries(knowledge)) {
        const row = addRow(table);
        addColumn(row, name);
        const input = addInput(null, null, values.join(", "));
        input.input.addEventListener("input", () => {
          knowledge[name] = input.input.value.split(",").map((a) => a.trim());
        });
        const v = addColumn(row, input.input);
        v.style.flex = "3";
        const b = addButton(null, "X", "red", () => {
          delete knowledge[name];
          renderKnowledge();
        });
        b.style.width = "12px";
        b.style.height = "12px";
        b.style.flex = "initial";
        b.style.verticalAlign = "middle";
        addColumn(row, b);
      }
      const modRow = addRow(table, true);
      const inputRaw = addInput(null, null, null, "Knowledge Name").input;

      const input = addColumn(modRow, inputRaw);
      input.style.flex = "4";
      const b = addButton(null, "Add", "blue", () => {
        knowledge[inputRaw.value] = [];
        renderKnowledge();
      });
      addColumn(modRow, b);
      startInputValidation(
        () =>
          isValidIdent(inputRaw.value) &&
          knowledge[inputRaw.value] === undefined,
        b,
        input,
      );
      addButton(table, "Save", "green", async () => {
        const values = {
          project: wArgs.project,
          knowledge,
        };
        await sendToServer("/update_knowledge", values);
      });
    }
    renderKnowledge();
  }

  /**
   * @param {HTMLElement} tabView - The current tab
   */
  function renderAnnotationView(tabView) {
    const { input: annotationNameInput } = addInput(
      tabView,
      "Annotation Name",
      null,
      "Annotation Name",
      "The name for the new annotation. This will end up as webdb='ANNOTATION_NAME' in the HTML and will be referenced by everything, that targets an html node. An annotation may only be alphanumeric, or '_' and at least 3 chars long",
    );
    const { select: annotateListSelect } = addSelect(
      tabView,
      "Annotation Count",
      ["single element", "simple list", "recursive list", "finish list"],
      "The amount of elements to annotate. Lists can be automatically annotated all at once. If you just need a list of elements close to each other, pick 'Simple List'. For more complex lists (multiple roles per actor for all actors), you may choose 'Recursive List'",
    );
    const { select: actionSelect } = addSelect(
      tabView,
      "Action",
      [
        "create data point",
        "create node",
        "only annotation",
        "manual data point",
        "manual node",
      ],
      "In many cases you can infer enough information from where you click to take over the work of creating a node/point just from the annotation name. That is what the 'create ..' options are for. Otherwise you can just annotate, or be forwarded to the create point/node dialog for manual configuration using the `manual` options",
    );

    const { select: dataTargetSelect, label: dataTargetSelectLabel } =
      addSelect(
        tabView,
        "Data Target",
        ["table_insert", "mtm_insert", "link", "node_ident", "ignore"],
        "This will create a new table column on whatever nodes table you select. Link also creates a page with the node name and links to that. Ident adds your selected text to the nodes idents",
      );

    const { select: dataTypeSelect, label: dataTypeSelectLabel } = addSelect(
      tabView,
      "Insertion Type",
      ["text", "int", "float"],
      "The datatype, that the new column should have in the db",
    );

    const projectConfig = wArgs.config[wArgs.project];
    const availableTables = ["-- New Table --"];
    for (const [name, info] of Object.entries(projectConfig.tables)) {
      if (info.type === "primary") {
        availableTables.push(name);
      }
    }
    const { select: tableSelect, label: tableSelectLabel } = addSelect(
      tabView,
      "Node Table",
      availableTables,
      "The table, that the data will be written into",
    );

    const { select: mtmSelect, label: mtmSelectLabel } = addSelect(
      tabView,
      "MtM Table",
      [],
      "The mtm table, that the data will be written into",
    );

    const { input: mtmInput, label: mtmInputLabel } = addInput(
      tabView,
      "Custom MtM table name",
      "",
      "The custom name of the mtm table to use",
    );
    function updateMtm() {
      const availableMtMTables = [
        "-- Auto MtM Table --",
        "-- New MtM Table --",
        "-- No MtM Table --",
      ];
      for (const [tableName, tableInfo] of Object.entries(
        projectConfig.tables,
      )) {
        if (tableInfo.type !== "many_to_many") {
          continue;
        }

        if (
          tableInfo.table_a === tableSelect.value ||
          tableInfo.table_b === tableSelect.value
        ) {
          availableMtMTables.push(tableName);
        }
      }
      addSelectOptions(mtmSelect, availableMtMTables);
    }
    tableSelect.addEventListener("change", updateMtm);
    updateMtm();

    const { select: valueListSelect, label: valueListSelectLabel } = addSelect(
      tabView,
      "Value List",
      ["no", "value_list", "string_split_list"],
      "If you are selecting a simple list, that only contains a single text string per node, this will correctly parse and insert that without having to manually add a point",
    );

    const { input: splitInput, label: splitLabel } = addInput(
      tabView,
      "List Split",
      ",",
      null,
      "The regex, that you want to split the nodes string value by",
    );

    function updateAuto() {
      hideElements(
        tableSelect,
        tableSelectLabel,
        dataTargetSelect,
        dataTargetSelectLabel,
        dataTypeSelect,
        dataTypeSelectLabel,
        mtmSelect,
        mtmSelectLabel,
        valueListSelect,
        valueListSelectLabel,
        splitInput,
        splitLabel,
        mtmInput,
        mtmInputLabel,
      );

      if (actionSelect.value === "create data point") {
        showElements(dataTargetSelect, dataTargetSelectLabel);
        if (
          dataTargetSelect.value === "table_insert" ||
          dataTargetSelect.value === "mtm_insert"
        ) {
          showElements(dataTypeSelect, dataTypeSelectLabel);
        }
      } else if (actionSelect.value === "create node") {
        showElements(
          tableSelect,
          tableSelectLabel,
          mtmSelect,
          mtmSelectLabel,
          valueListSelect,
          valueListSelectLabel,
        );
        if (valueListSelect.value === "string_split_list") {
          showElements(splitInput, splitLabel);
        }
        if (mtmSelect.value === "-- New MtM Table --") {
          showElements(mtmInput, mtmInputLabel);
        }
      }
    }
    mtmSelect.addEventListener("change", updateAuto);
    valueListSelect.addEventListener("change", updateAuto);
    dataTargetSelect.addEventListener("change", updateAuto);
    actionSelect.addEventListener("change", updateAuto);
    updateAuto();

    const annotateButton = addButton(tabView, "Annotate", "blue", () => {
      setTimeout(() => {
        annotate(
          annotationNameInput.value,
          annotateListSelect.value,
          actionSelect.value,
          tableSelect.value,
          dataTargetSelect.value,
          dataTypeSelect.value,
          mtmSelect.value,
          valueListSelect.value,
          splitInput.value,
          mtmInput.value,
        );
      }, 100);
    });

    startInputValidation(
      () => {
        if (!isValidIdent(annotationNameInput.value)) {
          return false;
        }
        if (
          actionSelect.value === "create node" &&
          mtmSelect.value === "-- New MtM Table --" &&
          !isValidIdent(mtmInput.value)
        ) {
          return false;
        }
        return true;
      },
      annotateButton,
      annotationNameInput,
    );
  }

  /**
   * @param {HTMLElement} tabView - The current tab
   */
  function renderTableView(tabView) {
    const projectConfig = wArgs.config[wArgs.project];
    const tables = projectConfig.tables ?? {};
    for (const [tableName, table] of Object.entries(tables)) {
      const columns = table.columns ?? {};
      const tableDiv = addTable(tabView);
      const tableInfoRow = addRow(tableDiv, true);
      addColumn(tableInfoRow, `Table: ${tableName}`);
      /** @type{string} */
      let tableType;
      if (table.type === "primary") {
        tableType = `Primary`;
      } else if (table.type === "secondary") {
        tableType = `Secondary: ${table.primary_table}`;
      } else if (table.type === "many_to_many") {
        tableType = `MTM: ${table.table_a}-${table.table_b}`;
      } else {
        throw new Error("Invalid table type");
      }
      addColumn(tableInfoRow, tableType);
      addColumn(tableInfoRow, table.has_timestamps ? "(Timestamped)" : "");
      const rmButton = addButton(null, "X", "red", async () => {
        await sendToServer("/remove_table", {
          project: wArgs.project,
          table: tableName,
        });
      });
      rmButton.style.width = "12px";
      rmButton.style.height = "12px";
      rmButton.style.flex = "initial";
      rmButton.style.verticalAlign = "middle";
      addColumn(tableInfoRow, rmButton);

      for (const [columnName, column] of Object.entries(columns)) {
        const columnRow = addRow(tableDiv);
        addColumn(columnRow, columnName);

        const columnType =
          typeof column === "string" ? column : (column?.type ?? "");

        addColumn(columnRow, capitalizeWords(columnType));

        const trackStatus =
          typeof column === "string" || !(column?.track_updates ?? false)
            ? "Untracked"
            : "Tracked";
        addColumn(columnRow, trackStatus);

        const rmButton = addButton(null, "X", "red", async () => {
          await sendToServer("/remove_column", {
            project: wArgs.project,
            table: tableName,
            column: columnName,
          });
        });
        rmButton.style.width = "12px";
        rmButton.style.height = "12px";
        rmButton.style.flex = "initial";
        rmButton.style.verticalAlign = "middle";
        addColumn(columnRow, rmButton);
      }

      const column = renderColumnCreation(tableDiv, tableName);
      column.classList.add("webdbRow");
    }
    addButton(tabView, "New Table", "blue", () => {
      moveToPage("tableAdd");
    });
  }

  /**
   * @param {HTMLElement} tabView - The current tab
   */
  function renderTableAdd(tabView) {
    const projectConfig = wArgs.config[wArgs.project];

    const { input: tableNameInput } = addInput(
      tabView,
      "Table Name:",
      "",
      "Table Name",
    );

    const { select: tableTypeSelect } = addSelect(tabView, "Table Type:", [
      "primary",
      "secondary",
      "many_to_many",
    ]);

    const { select: secondaryTableSelect, label: secondaryLabel } = addSelect(
      tabView,
      "Referenced Primary Table",
    );
    const { select: mtmTableASelect, label: mtmALabel } = addSelect(
      tabView,
      "MtM Primary Table 1",
    );
    const { select: mtmTableBSelect, label: mtmBLabel } = addSelect(
      tabView,
      "MtM Primary Table 2",
    );

    function updateTableSelectionVisibilities() {
      const selectedType = tableTypeSelect.value;
      hideElements(
        mtmTableASelect,
        mtmTableBSelect,
        mtmALabel,
        mtmBLabel,
        secondaryTableSelect,
        secondaryLabel,
      );

      if (selectedType === "secondary") {
        showElements(secondaryTableSelect, secondaryLabel);
        const tables = Object.keys(projectConfig.tables);
        addSelectOptions(secondaryTableSelect, tables);
      } else if (selectedType === "many_to_many") {
        showElements(mtmTableASelect, mtmTableBSelect, mtmALabel, mtmBLabel);
        [mtmTableASelect, mtmTableBSelect].forEach((select) => {
          const tables = Object.keys(wArgs.config[wArgs.project].tables);
          addSelectOptions(select, tables);
        });
      }
    }

    tableTypeSelect.addEventListener(
      "change",
      updateTableSelectionVisibilities,
    );
    updateTableSelectionVisibilities();

    const { checkbox: timestampsCheckbox } = addCheckBox(
      tabView,
      "Has Timestamps",
      false,
      "Tables will track the creating and last update of each row, when this is enabled",
    );
    const addTableFn = async () => {
      const tableType = tableTypeSelect.value;
      const tableName = tableNameInput.value.trim();
      const hasTimestamps = timestampsCheckbox.checked;

      let key = null;
      if (tableType === "primary") {
        key = {
          type: tableType,
        };
      } else if (tableType === "secondary") {
        key = {
          type: tableType,
          primary_table: secondaryTableSelect.value,
        };
      } else if (tableType === "many_to_many") {
        key = {
          type: tableType,
          table_a: mtmTableASelect.value,
          table_b: mtmTableBSelect.value,
        };
      } else {
        alert("Invalid table Table");
        return;
      }
      const values = {
        project: wArgs.project,
        table: tableName,
        key,
        has_timestamps: hasTimestamps,
      };
      await sendToServer("/add_table", values, "table");
    };

    const saveButton = addButton(tabView, "Create Table", "blue", addTableFn);
    function isValidInput() {
      const tableName = tableNameInput.value.trim();
      const tables = projectConfig.tables || {};

      if (!isValidIdent(tableName) || tables[tableName]) {
        return false;
      }

      if (tableTypeSelect.value === "secondary") {
        if (!tables[secondaryTableSelect.value]) {
          return false;
        }
      } else if (tableTypeSelect.value === "many to many") {
        if (!tables[mtmTableASelect.value] || !tables[mtmTableBSelect.value]) {
          return false;
        } else if (mtmTableASelect.value === mtmTableBSelect.value) {
          return false;
        }
      }
      return true;
    }
    startInputValidation(
      isValidInput,
      saveButton,
      tableNameInput,
      tableTypeSelect,
    );
  }

  /**
   * Starts the input validation process for the specified target element.
   * The button will be enabled or disabled based on the result of the validation function.
   * @param {Function} isValidationFunc - The function that checks the validity of the input.
   * @param {HTMLButtonElement} targetElement - The target button that will be enabled/disabled.
   * @param {...HTMLElement} elements - The elements that trigger the validation.
   */
  function startInputValidation(isValidationFunc, targetElement, ...elements) {
    targetElement.disabled = !isValidationFunc();
    elements.forEach((a) => {
      let typ = "input";
      if (a instanceof HTMLSelectElement) {
        typ = "change";
      }
      a.addEventListener(typ, () => {
        targetElement.disabled = !isValidationFunc();
      });
    });
  }
  /**
   * Will query the page for all available annotations and return them to the user.
   * TODO: Also look through existing project definition & xpaths
   * @returns {string[]} All the annotaions, that have been found
   */
  function getAllAnnotations() {
    /**@type {string[]} */
    const res = [];
    const docRes = document.querySelectorAll("[webdb]");
    for (const node of docRes) {
      const attribute = node.getAttribute("webdb");
      if (attribute && !res.includes(attribute)) {
        res.push(attribute);
      }
    }
    return res;
  }

  /**
   * Returns all nodes, that were annotated with @webdb='annotation'
   * @param {string} annotation
   * @returns {Element[]}
   */
  function getNodesForAnnotation(annotation) {
    const nodes = Array.from(
      document.querySelectorAll(`[webdb='${annotation}']`),
    );
    return nodes;
  }

  /**
   * @param {HTMLElement|null} tab - The current tab
   * @param {string} tableName - The name of the table we want to add a column to
   * @returns {HTMLDivElement}
   */
  function renderColumnCreation(tab, tableName) {
    const container = document.createElement("div");
    container.classList.add("webdbColumnAdd");

    const { input: columnNameInput } = addInput(
      container,
      null,
      null,
      "Column Name",
      null,
    );
    const { select: columnTypeSelect } = addSelect(
      container,
      null,
      ["text", "int", "float"],
      null,
    );
    const { checkbox: trackUpdatesCheckbox, container: checkContainer } =
      addCheckBox(
        container,
        "Tracked",
        false,
        "This will track updates to this row and notify the user, when it changes",
      );

    const addColumn = async () => {
      const columnName = columnNameInput.value.trim();
      const columnType = columnTypeSelect.value;
      const trackUpdates = trackUpdatesCheckbox.checked;

      if (columnName.length < 3) {
        return;
      }

      const values = {
        project: wArgs.project,
        table: tableName,
        column: columnName,
        column_type: columnType,
        track_updates: trackUpdates,
        is_unique: false,
      };
      await sendToServer("/add_column", values);
    };
    const button = addButton(container, "Add", "green", addColumn);

    // function isValid() {}

    [columnNameInput, columnTypeSelect, checkContainer, button].forEach((a) => {
      a.classList.add("webdbColumn");
    });
    if (tab) {
      tab.appendChild(container);
    }
    return container;
  }

  /**
   * @param {HTMLElement} tabView
   */
  function renderNodeView(tabView) {
    const projectConfig = wArgs.config[wArgs.project] ?? {};
    const nodes = projectConfig.sites[wArgs.site]?.[wArgs.page]?.nodes ?? {};

    for (const [nodeName, node] of Object.entries(nodes)) {
      const nodeTable = addTable(tabView);
      const nodeTitleRow = addRow(nodeTable, true);

      addColumn(nodeTitleRow, `Node: ${nodeName}`);

      const annotationName = node.position?.annotation
        ? `Annotation: ${node.position.annotation}`
        : "-- ROOT --";
      addColumn(nodeTitleRow, annotationName);

      let tableName = node.table?.name
        ? `Table: ${node.table.name}`
        : "No Table";
      if (node.table?.mtm?.table) {
        tableName += `, MTM: ${node.table?.mtm?.table}`;
      }
      addColumn(nodeTitleRow, tableName);

      const editButton = addColumn(
        nodeTitleRow,
        addButton(null, "Edit", "blue", () => {
          moveToPage("nodesAdd", { node: nodeName });
        }),
      );
      editButton.style.flex = "initial";

      const rmButton = addButton(null, "X", "red", async () => {
        await sendToServer("/remove_node", {
          project: wArgs.project,
          site: wArgs.site,
          page: wArgs.page,
          node: nodeName,
        });
      });
      rmButton.style.width = "12px";
      rmButton.style.height = "12px";
      rmButton.style.flex = "initial";
      rmButton.style.verticalAlign = "middle";
      addColumn(nodeTitleRow, rmButton);

      for (const [pointName, point] of Object.entries(node.points ?? {})) {
        const pointRow = addRow(nodeTable);
        addColumn(pointRow, `Point: ${pointName}`);
        addColumn(pointRow, `Annotation: ${point.annotation}`);
        addColumn(pointRow, point.extractor ?? "");
        const editButton = addButton(null, "Edit", "blue", () => {
          moveToPage("nodesPointAdd", { node: nodeName, point: pointName });
        });
        editButton.style.flex = "initial";
        addColumn(pointRow, editButton);

        const rmButton = addButton(null, "X", "red", async () => {
          await sendToServer("/remove_point", {
            project: wArgs.project,
            site: wArgs.site,
            page: wArgs.page,
            node: nodeName,
            data_point: pointName,
          });
        });
        rmButton.style.width = "12px";
        rmButton.style.height = "12px";
        rmButton.style.flex = "initial";
        rmButton.style.verticalAlign = "middle";
        addColumn(pointRow, rmButton);
      }
      addButton(nodeTable, "Add Point", "blue", () => {
        moveToPage("nodesPointAdd", { node: nodeName });
      });
    }
    addButton(tabView, "Add Node", "green", () => {
      moveToPage("nodesAdd");
    });
  }

  /**
   * @param {HTMLElement} tabView
   */
  function renderXpathView(tabView) {
    const projectConfig = wArgs.config[wArgs.project] ?? {};
    const site = projectConfig.sites[wArgs.site]?.[wArgs.page];

    const generateXpaths = async () => {
      const k = parseFloat(kinput.value);
      const values = {
        project: wArgs.project,
        site: wArgs.site,
        page: wArgs.page,
        html: document.documentElement.innerHTML,
        k: isNaN(k) ? 10 : k,
        complex: ccheckbox.checked,
      };
      await sendToServer("/xpaths", values);
    };

    const settingsBar = addRow(tabView);

    const xpButton = addButton(
      tabView,
      "Generate Xpaths",
      "red",
      generateXpaths,
    );

    const kcontainer = document.createElement("div");
    const { input: kinput } = addInput(
      kcontainer,
      "K",
      "10",
      "K",
      "This influences both the amount and quality of xpaths",
    );

    const { checkbox: ccheckbox, container: ccontainer } = addCheckBox(
      null,
      "Complex Xpaths",
      false,
      "This generates xpaths, that contain other xpaths within the [] parts. This will make the generation slower, but may generate better xpaths. This will automatically be enabled, if no xpaths could be found otherwise",
    );
    addColumn(settingsBar, kcontainer);
    addColumn(settingsBar, ccontainer);

    const annotations = getAllAnnotations();
    if (annotations.length > 0) {
      xpButton.title =
        "Send the current HTML page to the server to generate xpaths from the current annotations";
    } else {
      xpButton.disabled = true;
      xpButton.title =
        "You need at least one annotated node to generate xpaths";
    }
    tabView.appendChild(xpButton);

    for (const [xpathName, xpaths] of Object.entries(site?.xpaths ?? {})) {
      const table = addTable(tabView);
      const row = addRow(table, true);
      const name = addColumn(row, xpathName);
      name.style.flex = "5";
      addColumn(row, "r_score");
      addColumn(row, "f_score");
      addColumn(row, "");

      for (const xpath of typeof xpaths === "string"
        ? [{ xp: xpaths, r_score: 1.0, f_score: 1.0 }]
        : xpaths) {
        const row = addRow(table);
        const xp = addColumn(row, xpath.xp);
        xp.style.flex = "5";
        addColumn(row, xpath.r_score.toString());
        addColumn(row, xpath.f_score.toFixed(2));
        const setbutton = addButton(null, "Set", "blue", async () => {
          await sendToServer("/set_xpath", {
            project: wArgs.project,
            site: wArgs.site,
            page: wArgs.page,
            name: xpathName,
            value: xpath,
          });
        });
        addColumn(row, setbutton);
      }
    }
  }

  /**
   * @param {HTMLElement} tabView
   */
  function renderPageView(tabView) {
    const projectConfig = wArgs.config[wArgs.project] ?? {};
    const page = projectConfig.sites[wArgs.site]?.[wArgs.page];

    const { input: startInput } = addInput(
      tabView,
      "Starting Page",
      page.start ?? "",
      null,
      "When starting a crawling run, this value will be inserted into the que of pages to fetch. You can use '{KNOWLEDGE_NAME}' to refer to a list of values of defined knowledge. These values will then replace the '{}' parts",
    );

    addButton(tabView, "Use Current Page", "blue", () => {
      startInput.value = window.location.href;
    });

    const interactions = structuredClone(page?.interactions ?? []);
    const interactionTable = addTable(tabView);
    const infoTitle = addRow(interactionTable, true);
    addColumn(infoTitle, "Interactions");

    function updateInteractions() {
      interactionTable.innerHTML = "";
      for (let i = 0; i < interactions.length; i++) {
        const interaction = interactions[i];
        const interactionRow = addRow(interactionTable);
        const action = interaction.action;
        if (action === "click") {
          addColumn(interactionRow, "Click Element");
          const annotation = `Annotation: ${interaction.annotation}`;
          addColumn(interactionRow, annotation);
          const navigates = interaction?.navigates
            ? `Navigates to: ${interaction?.navigates}`
            : "";
          addColumn(interactionRow, navigates);
        } else if (action === "input") {
          addColumn(interactionRow, "Text Input");
          const annotation = `Annotation: ${interaction.annotation}`;
          addColumn(interactionRow, annotation);
          addColumn(interactionRow, `Data: ${interaction.data}`);
        } else if (action === "scroll_to_end") {
          addColumn(interactionRow, "Scroll To End Of Page");
        } else if (action === "send_current") {
          addColumn(interactionRow, "Force Extract Current Page");
        } else {
          interactionRow.textContent = `Unknown action${action}`;
        }
        const removeButton = addButton(null, "X", "red", () => {
          const pos = interactions.indexOf(interaction);
          if (pos !== -1) {
            interactions.splice(pos, 1);
          }
          updateInteractions();
        });
        const insertedButton = addColumn(interactionRow, removeButton);
        insertedButton.style.width = "12px";
        insertedButton.style.height = "12px";
        insertedButton.style.flex = "initial";
        insertedButton.style.verticalAlign = "middle";
      }

      const insertRow = addRow(interactionTable);
      const { select: typeSelection } = addSelect(insertRow, null, [
        "click",
        "input",
        "scroll_to_end",
        "send_current",
      ]);

      const annotations = getAllAnnotations();
      const { select: anSelection } = addSelect(insertRow, null, annotations);
      const site = projectConfig.sites[wArgs.site] ?? {};
      const pages = Object.keys(site);
      const { select: navSelection } = addSelect(insertRow, null, pages);
      const { input: dataInput } = addInput(insertRow, null, "");

      function updateBonus() {
        hideElements(anSelection, navSelection, dataInput);
        const action = typeSelection.value;
        if (action === "click") {
          showElements(anSelection, navSelection);
        } else if (action === "input") {
          showElements(anSelection, dataInput);
        }
      }

      typeSelection.addEventListener("change", updateBonus);
      updateBonus();

      const addInteractionButton = addButton(insertRow, "+", "blue", () => {
        const action = typeSelection.value;
        if (action === "click") {
          interactions.push({
            action,
            annotation: anSelection.value,
            navigates: navSelection.value,
          });
        } else if (action === "input") {
          interactions.push({
            action,
            annotation: anSelection.value,
            data: dataInput.value,
          });
        } else if (action === "send_current" || action === "scroll_to_end") {
          interactions.push({ action });
        }
        updateInteractions();
      });
      addInteractionButton.style.width = "12px";
      addInteractionButton.style.height = "12px";
      addInteractionButton.style.flex = "initial";
      addInteractionButton.style.verticalAlign = "middle";

      [typeSelection, anSelection, navSelection, dataInput].forEach((a) => {
        a.classList.add("webdbColumn");
      });
    }

    updateInteractions();

    addButton(tabView, "Save", "green", async () => {
      /**@type{null|string} */
      let start = startInput.value;
      if (start === "") {
        start = null;
      }

      const values = {
        project: wArgs.project,
        site: wArgs.site,
        page: wArgs.page,
        start,
        interactions,
      };
      await sendToServer("/update_page", values);
    });
  }

  /**
   *
   * @param {Element|null} node
   * @returns {number}
   */
  function nodePosition(node) {
    let position = 0;
    while (node !== null) {
      position += 1;
      node = node.previousElementSibling;
    }
    return position;
  }

  /**
   * @param {HTMLElement} start - The starting node in the list, that we want to start finding a list from
   * @param {(element: Node | null) => void} node_func - The function, that annotates the list
   * @param {boolean} recursive - Wether or not we want to stop after finding the first list
   */
  function annotateList(start, node_func, recursive) {
    let xpath = `child::${start.nodeName}`;
    for (let i = 0; i < start.classList.length; i++) {
      xpath += `[contains(@class, '${start.classList[i]}') ]`;
      break;
    }
    let parent = start.parentNode;
    while (parent) {
      const nodes = document.evaluate(
        xpath,
        parent,
        null,
        XPathResult.ORDERED_NODE_SNAPSHOT_TYPE,
        null,
      );
      for (let i = 0; i < nodes.snapshotLength; i++) {
        node_func(nodes.snapshotItem(i));
      }
      console.log(`SL: ${nodes.snapshotLength}`);
      if (recursive === false && nodes.snapshotLength > 1) {
        console.log(`broke on: ${xpath}`);
        break;
      }
      xpath = `child::${parent.nodeName}/${xpath}`;
      parent = parent.parentNode;
    }
  }

  /**
   * Hides the floating window and gives the over the annotation controls
   * @param {string} annotation - The annotation to use
   * @param {string} type - single/list/recursive
   * @param {string} action - node/point/manual
   * @param {string} autoTable
   * @param {string} autoType
   * @param {string} dataType
   * @param {string} autoMtm
   * @param {string} valueList
   * @param {string} split
   * @param {string} mtmInput
   */
  function annotate(
    annotation,
    type,
    action,
    autoTable,
    autoType,
    dataType,
    autoMtm,
    valueList,
    split,
    mtmInput,
  ) {
    hideElements(floatingWindow);

    /** @param {MouseEvent} event */
    function onMouseOver(event) {
      // Remove old modifiers
      document.querySelectorAll(".element-selector-hover").forEach((el) => {
        if (el.classList) {
          el.classList.remove("element-selector-hover");
        }
      });

      if (!(event.target instanceof HTMLElement)) {
        return;
      }

      if (["simple list", "recursive list"].includes(type)) {
        annotateList(
          event.target,
          (node) => {
            if (node instanceof HTMLElement) {
              node.classList.add("element-selector-hover");
            }
          },
          type === "recursive list",
        );
      } else if (type === "finish list") {
        const starts = Array.from(
          document.querySelectorAll(`[webdb='${annotation}']`),
        );
        if (event.target instanceof Element && !starts.includes(event.target)) {
          starts.push(event.target);
        }

        /**
         * @param {Element} node
         * @returns {Element[]}
         */
        function getAncestors(node) {
          const ancestors = [];
          while (node) {
            ancestors.push(node);
            if (node.parentElement instanceof Element) {
              node = node.parentElement;
            } else {
              break;
            }
          }
          return ancestors;
        }

        /**
         * @param {Element[]} elements
         * @returns {{ lca: Element|null, differingNodes: Element[][] }}
         */
        function findLeastCommonAncestor(elements) {
          if (!elements.length) {
            return { lca: null, differingNodes: [] };
          }

          const reversedAncestorLists = elements
            .map(getAncestors)
            .map((list) => list.reverse());

          /** @type{Element|null} */
          let lca = null;
          let commonDepth = 0;

          for (let i = 0; ; i++) {
            const levelNodes = reversedAncestorLists.map((list) => list[i]);
            const first = levelNodes[0];
            if (
              levelNodes.every((node) => node === first && node !== undefined)
            ) {
              lca = first;
              commonDepth = i + 1;
            } else {
              break;
            }
          }

          const differingNodes = reversedAncestorLists.map((list) =>
            list.slice(commonDepth),
          );
          for (const k of differingNodes) {
            k.pop();
          }

          return { lca, differingNodes };
        }

        /**
         *
         * @param {Element[]} nodes
         * @returns {{position: number|null, nodeName: string|null, classes: string[]} | null}
         */
        function findIdent(nodes) {
          const first = nodes[0];
          if (first === undefined || !nodes.every((a) => a !== undefined)) {
            return null;
          }
          /**
           * @type {{position: number|null, nodeName: string|null, classes: string[]}}
           */
          const ourIdent = {
            position: nodePosition(first),
            nodeName: first.nodeName,
            classes: [],
          };
          first.classList.values().forEach((a) => ourIdent.classes.push(a));
          for (let k = 1; k < nodes.length; k++) {
            const our = nodes[k];
            if (ourIdent.nodeName !== our.nodeName) {
              ourIdent.nodeName = null;
            }
            ourIdent.classes = ourIdent.classes.filter((a) =>
              our.classList.contains(a),
            );
            const pos = nodePosition(our);
            if (ourIdent.position !== pos) {
              ourIdent.position = null;
            }
          }
          return ourIdent;
        }

        const { lca, differingNodes } = findLeastCommonAncestor(starts);
        if (lca === null) {
          alert("Could not find a common ancestor");
          return;
        }

        let xpath = "";
        for (let i = 0; ; i++) {
          const nodes = differingNodes.map((list) => list[i]);
          const ident = findIdent(nodes);

          if (ident === null) {
            break;
          }

          if (xpath.length !== 0) {
            xpath += "/";
          }
          if (ident.position) {
            xpath += `child::*[${ident.position}]`;
          } else if (ident.classes.length !== 0) {
            xpath += `child::${ident.nodeName ?? "*"}`;
            for (const c of ident.classes) {
              xpath += `[contains(@class, '${c}')]`;
            }
          } else if (ident.nodeName) {
            xpath += `child::${ident.nodeName}`;
          } else {
            xpath += `child::*`;
          }
        }

        const ident = findIdent(starts);

        if (ident !== null) {
          if (xpath.length !== 0) {
            xpath += "/";
          }
          if (ident.position) {
            xpath += `descendant::*[${ident.position}]`;
          } else if (ident.classes.length !== 0) {
            xpath += `descendant::${ident.nodeName ?? "*"}`;
            for (const c of ident.classes) {
              xpath += `[contains(@class, '${c}')]`;
            }
          } else if (ident.nodeName) {
            xpath += `descendant::${ident.nodeName}`;
          } else {
            xpath += `descendant::*`;
          }
        } else {
          alert("bad starts");
        }

        console.log(xpath);
        const nodes = document.evaluate(
          xpath,
          lca,
          null,
          XPathResult.ORDERED_NODE_SNAPSHOT_TYPE,
          null,
        );

        for (let i = 0; i < nodes.snapshotLength; i++) {
          const node = nodes.snapshotItem(i);
          if (node instanceof Element) {
            node.classList.add("element-selector-hover");
          }
        }
      } else {
        event.target.classList.add("element-selector-hover");
      }
    }
    document.addEventListener("mouseover", onMouseOver);

    /** @param {MouseEvent} event */
    async function onMouseClick(event) {
      event.preventDefault();
      event.stopPropagation();
      event.stopImmediatePropagation();

      const selected = document.querySelectorAll(".element-selector-hover");
      selected.forEach((el) => {
        if (el instanceof HTMLElement) {
          el.classList.remove("element-selector-hover");
          if (!el.attributes.getNamedItem("webdb")) {
            el.setAttribute("webdb", annotation);
          }
        }
      });

      showElements(floatingWindow);
      document.removeEventListener("mouseover", onMouseOver);

      const projectConfig = wArgs.config[wArgs.project] ?? {};
      const nodes = projectConfig.sites[wArgs.site]?.[wArgs.page]?.nodes ?? {};

      /**
       * @returns {{name: string, info: NodeInfo} | null}
       */
      function findParentNode() {
        /**@type {{name: string, info: NodeInfo} | null} */
        let result = null;

        selected.forEach((a) => {
          while (a.parentNode instanceof HTMLElement) {
            const attr = a.parentNode.getAttribute("webdb");
            a = a.parentNode;
            if (!attr) {
              continue;
            }
            for (const [name, info] of Object.entries(nodes)) {
              if (info.position?.annotation === attr) {
                result = { name, info };
                return;
              }
            }
          }
        });

        if (!result) {
          for (const [name, info] of Object.entries(nodes)) {
            if (!info.position) {
              result = { name, info };
              break;
            }
          }
        }
        return result;
      }
      function resetAnnotation() {
        selected.forEach((el) => {
          if (el instanceof HTMLElement) {
            el.removeAttribute("webdb");
          }
        });
      }

      if (action === "manual data point") {
        moveToPage("nodesPointAdd", { annotation });
      } else if (action === "manual node") {
        moveToPage("nodesAdd", { annotation });
      } else if (action === "only annotation") {
        moveToPage("annotation");
      } else if (action === "create node") {
        const parent = findParentNode();
        if (!parent) {
          alert("Could not find a valid parent for this node");
          resetAnnotation();
          moveToPage("annotation");
          return;
        }

        const tableName = autoTable.startsWith("--") ? annotation : autoTable;

        /**@type{string|undefined} */
        let mtmName = autoMtm;
        if (mtmName === "-- Auto MtM Table --") {
          if (!parent.info.table) {
            mtmName = undefined;
          } else {
            mtmName = `${annotation}2${parent.info.table.name}`;
          }
        } else if (mtmName === "-- New MtM Table --") {
          mtmName = mtmInput;
        } else if (mtmName === "-- No MtM Table --") {
          mtmName = undefined;
        }
        // TODO: Figure out parent
        const addNodeValues = {
          project: wArgs.project,
          site: wArgs.site,
          page: wArgs.page,
          annotation,

          parent: parent.name,
          table: tableName,
          mtm: mtmName,
          value_list: valueList,
          split: valueList === "string_split_list" ? split : undefined,
        };
        const res = await sendToServer("/add_auto_node", addNodeValues, null);
        if (res === null) {
          resetAnnotation();
        }
        moveToPage("annotation");
      } else if (action === "create data point") {
        const parent = findParentNode();
        if (!parent) {
          alert("Could not find a valid parent for this point");
          resetAnnotation();
          moveToPage("annotation");
          return;
        }
        const res = await sendToServer(
          "/add_auto_point",
          {
            project: wArgs.project,
            site: wArgs.site,
            page: wArgs.page,
            typ:
              autoType === "table_insert" || autoType === "mtm_insert"
                ? {
                    type: autoType,
                    data: dataType,
                  }
                : { type: autoType },
            annotation,
            parent: parent.name,
          },
          null,
        );

        if (res === null) {
          resetAnnotation();
        }

        moveToPage("annotation");
      } else {
        alert("Invalid return type");
      }
    }

    document.addEventListener("click", onMouseClick, {
      once: true,
      capture: true,
    });

    document.addEventListener(
      "keydown",
      (event) => {
        if (event.key === "Escape") {
          showElements(floatingWindow);
          document.removeEventListener("mouseover", onMouseOver);
          document.removeEventListener("click", onMouseClick);
          document.querySelectorAll(".element-selector-hover").forEach((el) => {
            if (el.classList) {
              el.classList.remove("element-selector-hover");
            }
          });
        }
      },
      {
        once: true,
      },
    );
  }

  document.body.appendChild(floatingWindow);
}
